"""Calculate the hash of files."""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, ClassVar, Iterable, Optional

if TYPE_CHECKING:
    import hashlib

    from _typeshed import StrPath


class FileHash:
    """Wrapper for hashlib to easily calculate file hashes.

    Attributes:
        DEFAULT_CHUNK_SIZE: Default chunk size if not defined.

    Note:
        Does not support algorithms with variable length digests (e.g. SHAKE).

    """

    DEFAULT_CHUNK_SIZE: ClassVar[int] = (
        1024 * 10_000_000  # 10mb - number of bytes in each read operation
    )

    def __init__(
        self, hash_alg: "hashlib._Hash", *, chunk_size: int = DEFAULT_CHUNK_SIZE
    ) -> None:
        """Instantiate class.

        Args:
            hash_alg: Instance of a hashlib algorithm.
            chunk_size: When reading a file, it will be read this many bytes at
                a time. Larger values are more time efficient while smaller
                values or more memory efficient.

        """
        self._hash = hash_alg  # protected to discourage direct access
        self.chunk_size = chunk_size

    @property
    def digest(self) -> bytes:
        """Digest of the data hashed so far.

        Returns:
            This is a bytes object of size ``digest_size`` which may contain
            bytes in the whole range from 0 to 255.

        """
        return self._hash.digest()

    @property
    def digest_size(self) -> int:
        """Size of the resulting hash in bytes."""
        return self._hash.digest_size

    @property
    def hexdigest(self) -> str:
        """Digest of the data hashed so far.

        Returns:
            String object that is double the length of ``digest`` and contains
            only hexadecimal digits.

        """
        return self._hash.hexdigest()

    def add_file(self, file_path: StrPath) -> None:
        """Add file contents to the hash.

        Args:
            file_path: Path of the file to add.

        """
        with open(file_path, "rb") as stream:
            # python 3.7 compatable version of `while chunk := buf.read(read_size):`
            chunk = stream.read(self.chunk_size)  # seed chunk with initial value
            while chunk:
                self._hash.update(chunk)
                chunk = stream.read(self.chunk_size)  # read in new chunk

    def add_file_name(
        self,
        file_path: StrPath,
        *,
        end_character: str = "\0",
        relative_to: Optional[StrPath] = None,
    ) -> None:
        """Add file name to the hash. This includes the path.

        Args:
            file_path: Path of the file to add. The full path (or relative) is
                included when adding it to the hash. This is not resolved prior
                to use. It is used as-is unless another argument acts up it.
            end_character: Character that will be added to the end of the file_path.
                This can be an empty string.
            relative_to: Optionally, convert the file_path to path relative to
                this one. It is recommended that both paths be absolute.

        """
        self._hash.update(
            (
                str(
                    Path(file_path).relative_to(relative_to)
                    if relative_to
                    else Path(file_path)
                )
                + end_character
            ).encode()
        )

    def add_files(
        self,
        file_paths: Iterable[StrPath],
        *,
        relative_to: Optional[StrPath] = None,
    ) -> None:
        """Add files to the hash.

        Args:
            file_paths: Paths of the files to add. The full path (or relative) is
                included when adding it to the hash. This is not resolved prior
                to use. It is used as-is unless another argument acts up it.
            relative_to: Optionally, convert the file_path to path relative to
                this one. It is recommended that both paths be absolute.

        """
        for fp in file_paths:
            self.add_file_name(fp, relative_to=relative_to)
            self.add_file(fp)
            # end of file contents; only necessary with multiple files
            self._hash.update("\0".encode())



"""CFNgin blueprint representing raw template module."""
from __future__ import annotations

import hashlib
import json
import logging
import os
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

from jinja2 import Environment, FileSystemLoader

from ...compat import cached_property
from ..exceptions import InvalidConfig, UnresolvedBlueprintVariable
from ..utils import parse_cloudformation_template
from .base import Blueprint

if TYPE_CHECKING:
    from ...context import CfnginContext
    from ...variables import Variable

LOGGER = logging.getLogger(__name__)


def get_template_path(file_path: Path) -> Optional[Path]:
    """Find raw template in working directory or in sys.path.

    template_path from config may refer to templates co-located with the CFNgin
    config, or files in remote package_sources. Here, we emulate python module
    loading to find the path to the template.

    Args:
        filename: Template path.

    Returns:
        Path to file, or None if no file found

    """
    if file_path.is_file():
        return file_path
    for i in sys.path:
        test_path = Path(i) / file_path.name
        if test_path.is_file():
            return test_path
    return None


def resolve_variable(provided_variable: Optional[Variable], blueprint_name: str) -> Any:
    """Resolve a provided variable value against the variable definition.

    This acts as a subset of resolve_variable logic in the base module, leaving
    out everything that doesn't apply to CFN parameters.

    Args:
        provided_variable: The variable value provided to the blueprint.
        blueprint_name: The name of the blueprint that the variable is
            being applied to.

    Raises:
        UnresolvedBlueprintVariable: Raised when the provided variable is
            not already resolved.

    """
    value = None
    if provided_variable:
        if not provided_variable.resolved:
            raise UnresolvedBlueprintVariable(blueprint_name, provided_variable)

        value = provided_variable.value

    return value


class RawTemplateBlueprint(Blueprint):  # pylint: disable=abstract-method
    """Blueprint class for blueprints auto-generated from raw templates.

    Attributes:
        context: CFNgin context object.
        description: The description of the CloudFormation template that will
            be generated by this Blueprint.
        mappings: CloudFormation Mappings to be added to the template during the
            rendering process.
        name: Name of the Stack that will be created by the Blueprint.
        raw_template_path: Path to the raw CloudFormation template file.

    """

    raw_template_path: Path

    def __init__(  # pylint: disable=super-init-not-called
        self,
        name: str,
        context: CfnginContext,
        *,
        description: Optional[str] = None,
        mappings: Optional[Dict[str, Any]] = None,
        raw_template_path: Path,
        **_: Any,
    ) -> None:
        """Instantiate class.

        .. versionchanged:: 2.0.0
            Class only takes 2 positional arguments.
            The rest are now keyword arguments.

        """
        self._rendered = None
        self._resolved_variables = None
        self._version = None
        self.context = context
        self.description = description
        self.mappings = mappings
        self.name = name
        self.raw_template_path = raw_template_path

    @property
    def output_definitions(self) -> Dict[str, Dict[str, Any]]:
        """Get the output definitions.

        .. versionadded:: 2.0.0

        Returns:
            Output definitions. Keys are output names, the values are dicts
            containing key/values for various output properties.

        """
        return self.to_dict().get("Outputs", {})

    @cached_property
    def parameter_definitions(self) -> Dict[str, Any]:
        """Get the parameter definitions to submit to CloudFormation.

        .. versionadded:: 2.0.0

        Returns:
            Parameter definitions. Keys are parameter names, the values are dicts
            containing key/values for various parameter properties.

        """
        return self.to_dict().get("Parameters", {})

    @cached_property
    def parameter_values(self) -> Dict[str, Union[List[Any], str]]:
        """Return a dict of variables with type :class:`~runway.cfngin.blueprints.variables.types.CFNType`.

        .. versionadded:: 2.0.0

        Returns:
            Variables that need to be submitted as CloudFormation Parameters.
            Will be a dictionary of ``<parameter name>: <parameter value>``.

        """  # noqa
        return self._resolved_variables or {}

    @property
    def rendered(self) -> str:
        """Return (generating first if needed) rendered template."""
        if not self._rendered:
            template_path = get_template_path(self.raw_template_path)
            if template_path:
                if len(os.path.splitext(template_path)) == 2 and (
                    os.path.splitext(template_path)[1] == ".j2"
                ):
                    self._rendered = (
                        Environment(
                            loader=FileSystemLoader(
                                searchpath=os.path.dirname(template_path)
                            )
                        )
                        .get_template(os.path.basename(template_path))
                        .render(
                            context=self.context,
                            mappings=self.mappings,
                            name=self.name,
                            variables=self._resolved_variables,
                        )
                    )
                else:
                    with open(template_path, "r", encoding="utf-8") as template:
                        self._rendered = template.read()
            else:
                raise InvalidConfig(f"Could not find template {self.raw_template_path}")
            # clear cached properties that rely on this property
            self._del_cached_property("parameter_definitions")

        return self._rendered

    @property
    def requires_change_set(self) -> bool:
        """Return True if the underlying template has transforms."""
        return bool("Transform" in self.to_dict())

    @property
    def version(self) -> str:
        """Return (generating first if needed) version hash."""
        if not self._version:
            self._version = hashlib.md5(self.rendered.encode()).hexdigest()[:8]
        return self._version

    def to_dict(self) -> Dict[str, Any]:
        """Return the template as a python dictionary.

        Returns:
            dict: the loaded template as a python dictionary

        """
        return parse_cloudformation_template(self.rendered)

    def to_json(self, variables: Optional[Dict[str, Any]] = None) -> str:
        """Return the template in JSON.

        Args:
            variables: Unused in this subclass (variables won't affect the template).

        """
        # load -> dumps will produce json from json or yaml templates
        return json.dumps(self.to_dict(), sort_keys=True, indent=4)

    def render_template(self) -> Tuple[str, str]:
        """Load template and generate its md5 hash."""
        return (self.version, self.rendered)

    def resolve_variables(self, provided_variables: List[Variable]) -> None:
        """Resolve the values of the blueprint variables.

        This will resolve the values of the template parameters with values
        from the env file, the config, and any lookups resolved. The
        resolution is run twice, in case the blueprint is jinja2 templated
        and requires provided variables to render.

        Args:
            provided_variables: List of provided variables.

        """
        # Pass 1 to set resolved_variables to provided variables
        self._resolved_variables = {}
        variable_dict = {var.name: var for var in provided_variables}
        for var_name, _var_def in variable_dict.items():
            value = resolve_variable(variable_dict.get(var_name), self.name)
            if value is not None:
                self._resolved_variables[var_name] = value

        # Pass 2 to render the blueprint and set resolved_variables according
        # to defined variables
        # save a copy of param defs before clearing resolved var dict
        defined_variables = self.parameter_definitions.copy()
        self._resolved_variables = {}
        variable_dict = {var.name: var for var in provided_variables}
        for var_name, _var_def in defined_variables.items():
            value = resolve_variable(variable_dict.get(var_name), self.name)
            if value is not None:
                self._resolved_variables[var_name] = value
        # clear cached properties that rely on the property set by this
        self._del_cached_property("parameter_values")



"""Source code."""
from __future__ import annotations

import hashlib
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Iterator, List, Optional, Sequence, Union

import igittigitt

from runway.compat import cached_property
from runway.utils import FileHash

if TYPE_CHECKING:
    from _typeshed import StrPath

LOGGER = logging.getLogger(__name__)


class SourceCode:
    """Source code iterable."""

    gitignore_filter: igittigitt.IgnoreParser
    """Filter to use when zipping dependencies.
    If file/folder matches the filter, it should be ignored.

    """

    project_root: Path
    """Top-level directory containing the project metadata files and source code root directory.
    The value can be the same as ``root_directory``.
    If it is not, it must be a parent of ``root_direcotry``.

    """

    root_directory: Path
    """The root directory containing the source code."""

    def __init__(
        self,
        root_directory: StrPath,
        *,
        gitignore_filter: Optional[igittigitt.IgnoreParser] = None,
        include_files_in_hash: Optional[Sequence[Path]] = None,
        project_root: Optional[StrPath] = None,
    ) -> None:
        """Instantiate class.

        Args:
            root_directory: The root directory containing the source code.
            gitignore_filter: Object that has been pre-populated with
                rules/patterns to determine if a file should be ignored.
            include_files_in_hash: Files that should be included in hash
                calculation even if they are filtered by gitignore (e.g.
                ``poetry.lock``).
            project_root: Optional project root if the source code is located
                within a larger project. This should only be used if the
                contents of value of ``include_files_in_hash`` contains paths
                that exist outside of the root directory. If this is provided,
                it must be a parent of the root directory.

        """
        self._include_files_in_hash = include_files_in_hash or []
        self.gitignore_filter = gitignore_filter or igittigitt.IgnoreParser()
        self.root_directory = (
            root_directory if isinstance(root_directory, Path) else Path(root_directory)
        )
        self.project_root = (  # defaults to root_directory if project_root not provided
            project_root
            if isinstance(project_root, Path)
            else (Path(project_root) if project_root else self.root_directory)
        )

        if not gitignore_filter:
            self.gitignore_filter.parse_rule_files(self.root_directory)
            self.gitignore_filter.add_rule(".git/", self.root_directory)
            self.gitignore_filter.add_rule(".gitignore", self.root_directory)

    @cached_property
    def md5_hash(self) -> str:
        """Calculate the md5 hash of the directory contents.

        This can be resource intensive depending on the size of the project.

        """
        sorted_files = list(self.sorted())
        for include_file in self._include_files_in_hash:
            if include_file not in sorted_files:
                sorted_files.append(include_file)
        file_hash = FileHash(hashlib.md5())
        file_hash.add_files(sorted(sorted_files), relative_to=self.project_root)
        return file_hash.hexdigest

    def add_filter_rule(self, pattern: str) -> None:
        """Add rule to ignore filter.

        Args:
            pattern: The gitignore pattern to add to the filter.

        """
        self.gitignore_filter.add_rule(pattern=pattern, base_path=self.root_directory)

    def sorted(self, *, reverse: bool = False) -> List[Path]:
        """Sorted list of source code files.

        Args:
            reverse: Sort the list in reverse.

        Returns:
            Sorted list of source code files excluding those that match the
            ignore filter.

        """
        return sorted(self, reverse=reverse)

    def __eq__(self, other: object) -> bool:
        """Compare if self is equal to another object."""
        if isinstance(other, SourceCode):
            return self.root_directory == other.root_directory
        return False

    def __fspath__(self) -> Union[str, bytes]:
        """Return the file system path representation of the object."""
        return str(self.root_directory)

    def __iter__(self) -> Iterator[Path]:
        """Iterate over the source code files.

        Yields:
            Files that do not match the ignore filter. Order in arbitrary.

        """
        for child in self.root_directory.rglob("*"):
            if child.is_dir():
                continue  # ignore directories
            if self.gitignore_filter.match(child):
                continue  # ignore files that match the filter
            yield child

    def __str__(self) -> str:
        """Return the string representation of the object."""
        return str(self.root_directory)

    def __truediv__(self, other: StrPath) -> Path:
        """Create a new path object from source code's root directory."""
        return self.root_directory / other



"""Runway test definition models."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Union, cast

from pydantic import Extra, Field, validator
from typing_extensions import Literal

from .. import utils
from ..base import ConfigProperty

if TYPE_CHECKING:
    from typing import Callable

ValidRunwayTestTypeValues = Literal["cfn-lint", "script", "yamllint"]


class RunwayTestDefinitionModel(ConfigProperty):
    """Model for a Runway test definition."""

    args: Union[Dict[str, Any], ConfigProperty, str] = Field(
        default={},
        title="Arguments",
        description="Arguments to be passed to the test. Support varies by test type.",
    )
    name: str = Field(default="test-name", description="Name of the test.")
    required: Union[bool, str] = Field(
        default=False,
        description="Whether the test must pass for subsequent tests to be run.",
    )
    type: ValidRunwayTestTypeValues

    class Config(ConfigProperty.Config):
        """Model configuration."""

        schema_extra = {
            "description": "Tests that can be run via the 'test' command.",
        }
        title = "Runway Test Definition"
        use_enum_values = True

    def __new__(cls, **kwargs: Any) -> RunwayTestDefinitionModel:
        """Create a new instance of a class.

        Returns:
            Correct subclass of RunwayTestDefinition for the given data.

        """
        test_type = kwargs.get("type")
        if cls is RunwayTestDefinitionModel:
            if test_type == "cfn-lint":
                return super().__new__(CfnLintRunwayTestDefinitionModel)
            if test_type == "script":
                return super().__new__(ScriptRunwayTestDefinitionModel)
            if test_type == "yamllint":
                return super().__new__(YamlLintRunwayTestDefinitionModel)
        return super().__new__(cls)

    # TODO add regex to schema
    _validate_string_is_lookup = cast(
        "classmethod[Callable[..., Any]]",
        validator("args", "required", allow_reuse=True, pre=True)(
            utils.validate_string_is_lookup
        ),
    )


class CfnLintRunwayTestArgs(ConfigProperty):
    """Model for the args of a cfn-lint test."""

    cli_args: Union[List[str], str] = Field(
        default=[],
        title="CLI Arguments",
        description="Array of arguments to pass to the cfn-lint CLI.",
    )

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.forbid
        schema_extra = {
            "description": "Arguments supported by the cfn-lint test.",
        }
        title = "cfn-lint Runway Test Arguments"

    # TODO add regex to schema
    _validate_string_is_lookup = cast(
        "classmethod[Callable[..., Any]]",
        validator("cli_args", allow_reuse=True, pre=True)(
            utils.validate_string_is_lookup
        ),
    )


class CfnLintRunwayTestDefinitionModel(RunwayTestDefinitionModel):
    """Model for a cfn-lint test definition."""

    args: CfnLintRunwayTestArgs = Field(
        default=CfnLintRunwayTestArgs(),
        title="Arguments",
        description="Arguments to be passed to the test.",
    )
    name: str = Field(default="cfn-lint", description="Name of the test.")
    required: Union[bool, str] = Field(
        default=False,
        description="Whether the test must pass for subsequent tests to be run.",
    )
    type: Literal["cfn-lint"] = Field(
        default="cfn-lint", description="The type of test to run."
    )

    class Config(RunwayTestDefinitionModel.Config):
        """Model configuration."""

        schema_extra = {
            "description": "Test using cfn-lint.",
        }
        title = "cfn-lint Test"


class ScriptRunwayTestArgs(ConfigProperty):
    """Model for the args of a script test."""

    commands: Union[List[str], str] = Field(
        default=[], description="Array of commands that will be run for this test."
    )

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.forbid
        schema_extra = {
            "description": "Arguments supported by the script test.",
        }
        title = "Script Runway Test Arguments"

    # TODO add regex to schema
    _validate_string_is_lookup = cast(
        "classmethod[Callable[..., Any]]",
        validator("commands", allow_reuse=True, pre=True)(
            utils.validate_string_is_lookup
        ),
    )


class ScriptRunwayTestDefinitionModel(RunwayTestDefinitionModel):
    """Model for a script test definition."""

    args: ScriptRunwayTestArgs = Field(
        default=ScriptRunwayTestArgs(),
        title="Arguments",
        description="Arguments to be passed to the test.",
    )
    name: str = Field(default="script", description="Name of the test.")
    required: Union[bool, str] = Field(
        default=False,
        description="Whether the test must pass for subsequent tests to be run.",
    )
    type: Literal["script"] = Field(
        default="script", description="The type of test to run."
    )

    class Config(RunwayTestDefinitionModel.Config):
        """Model configuration."""

        schema_extra = {
            "description": "Test using a custom script.",
        }
        title = "Script Test"


class YamlLintRunwayTestDefinitionModel(RunwayTestDefinitionModel):
    """Model for a yamllint test definition."""

    name: str = Field(default="yamllint", description="Name of the test.")
    required: Union[bool, str] = Field(
        default=False,
        description="Whether the test must pass for subsequent tests to be run.",
    )
    type: Literal["yamllint"] = Field(
        default="yamllint", description="The type of test to run."
    )

    class Config(RunwayTestDefinitionModel.Config):
        """Model configuration."""

        schema_extra = {
            "description": "Test using yamllint.",
        }
        title = "yamllint Test"



"""``runway gen-sample sls-py`` command."""
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import click

from ... import options
from .utils import TEMPLATES, convert_gitignore, copy_sample

if TYPE_CHECKING:
    from ...._logging import RunwayLogger

LOGGER = cast("RunwayLogger", logging.getLogger(__name__.replace("._", ".")))


@click.command("sls-py", short_help="sls + python (sampleapp.sls)")
@options.debug
@options.no_color
@options.verbose
@click.pass_context
def sls_py(ctx: click.Context, **_: Any) -> None:
    """Generate a sample Serverless project using python."""
    src = TEMPLATES / "sls-py"
    dest = Path.cwd() / "sampleapp.sls"

    copy_sample(ctx, src, dest)
    convert_gitignore(dest / "_gitignore")

    LOGGER.success("Sample Serverless module created at %s", dest)
    LOGGER.notice(
        "To finish it's setup, change to the %s directory and "
        'execute "npm install" to generate it\'s lockfile.',
        dest,
    )



"""Docker image push hook.

Replicates the functionality of the ``docker image push`` CLI command.

"""
# pylint: disable=no-self-argument,no-self-use
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from pydantic import Field, validator

from .....utils import BaseModel
from ..data_models import (
    DockerImage,
    ElasticContainerRegistry,
    ElasticContainerRegistryRepository,
)
from ..hook_data import DockerHookData

if TYPE_CHECKING:
    from .....context import CfnginContext

LOGGER = logging.getLogger(__name__.replace("._", "."))


class ImagePushArgs(BaseModel):
    """Args passed to image.push."""

    _ctx: Optional[CfnginContext] = Field(default=None, alias="context", export=False)

    ecr_repo: Optional[ElasticContainerRegistryRepository] = None  # depends on _ctx
    """AWS Elastic Container Registry repository information.
    Providing this will automatically construct the repo URI.
    If provided, do not provide ``repo``.

    If using a private registry, only ``repo_name`` is required.
    If using a public registry, ``repo_name`` and ``registry_alias``.

    """

    image: Optional[DockerImage] = None
    """Image to push."""

    repo: Optional[str] = None  # depends on ecr_repo & image
    """URI of a non Docker Hub repository where the image will be stored."""

    tags: List[str] = []  # depends on image
    """List of tags to push."""

    @validator("ecr_repo", pre=True, allow_reuse=True)
    def _set_ecr_repo(cls, v: Any, values: Dict[str, Any]) -> Any:
        """Set the value of ``ecr_repo``."""
        if v and isinstance(v, dict):
            return ElasticContainerRegistryRepository.parse_obj(
                {
                    "repo_name": v.get("repo_name"),
                    "registry": ElasticContainerRegistry.parse_obj(
                        {
                            "account_id": v.get("account_id"),
                            "alias": v.get("registry_alias"),
                            "aws_region": v.get("aws_region"),
                            "context": values.get("context"),
                        }
                    ),
                }
            )
        return v

    @validator("repo", pre=True, always=True, allow_reuse=True)
    def _set_repo(cls, v: Optional[str], values: Dict[str, Any]) -> Optional[str]:
        """Set the value of ``repo``."""
        if v:
            return v

        image: Optional[DockerImage] = values.get("image")
        if image:
            return image.repo

        ecr_repo: Optional[ElasticContainerRegistryRepository] = values.get("ecr_repo")
        if ecr_repo:
            return ecr_repo.fqn

        return None

    @validator("tags", pre=True, always=True, allow_reuse=True)
    def _set_tags(cls, v: List[str], values: Dict[str, Any]) -> List[str]:
        """Set the value of ``tags``."""
        if v:
            return v

        image: Optional[DockerImage] = values.get("image")
        if image:
            return image.tags

        return ["latest"]


def push(*, context: CfnginContext, **kwargs: Any) -> DockerHookData:
    """Docker image push hook.

    Replicates the functionality of ``docker image push`` CLI command.

    kwargs are parsed by :class:`~runway.cfngin.hooks.docker.image.ImagePushArgs`.

    """
    args = ImagePushArgs.parse_obj({"context": context, **kwargs})
    docker_hook_data = DockerHookData.from_cfngin_context(context)
    LOGGER.info("pushing image %s...", args.repo)
    for tag in args.tags:
        docker_hook_data.client.images.push(repository=args.repo, tag=tag)
        LOGGER.info("successfully pushed image %s:%s", args.repo, tag)
    return docker_hook_data.update_context(context)



"""Execute the AWS CLI update-kubeconfig to generate your kubectl config."""
from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import TYPE_CHECKING, Any, List

from runway.cfngin.lookups.handlers.output import OutputLookup

if TYPE_CHECKING:
    from runway.context import CfnginContext

LOGGER = logging.getLogger(__name__)


def assumed_role_to_principle(assumed_role_arn: str) -> str:
    """Return role ARN from assumed role ARN."""
    arn_split = assumed_role_arn.split(":")
    arn_split[2] = "iam"
    base_arn = ":".join(arn_split[:5]) + ":role/"
    return base_arn + assumed_role_arn.split("/")[1]


def get_principal_arn(context: CfnginContext) -> str:
    """Return ARN of current session principle."""
    # looking up caller identity
    session = context.get_session()
    sts_client = session.client("sts")
    caller_identity_arn = sts_client.get_caller_identity().get("Arn", "")
    if caller_identity_arn.split(":")[2] == "iam" and (
        caller_identity_arn.split(":")[5].startswith("user/")
    ):
        return caller_identity_arn  # user arn
    return assumed_role_to_principle(caller_identity_arn)


def generate(
    context: CfnginContext, *, filename: str, path: List[str], stack: str, **_: Any
):
    """Generate an EKS auth_map for worker connection.

    Args:
        context: Context object.
        filename: Name of the file.
        path: Path to the file.
        stack: Stack definition.

    Returns:
        boolean for whether or not the hook succeeded

    """
    overlay_path = Path(*path)
    file_path = overlay_path / filename
    if os.path.exists(filename):
        LOGGER.info("%s file present; skipping initial creation", file_path)
        return True
    LOGGER.info("Creating auth_map at %s", file_path)
    overlay_path.mkdir(parents=True, exist_ok=True)
    principal_arn = get_principal_arn(context)
    node_instancerole_arn = OutputLookup.handle(
        f"{stack}::NodeInstanceRoleArn", context=context
    )
    aws_authmap_template = (Path(__file__).parent / "aws-auth-cm.yaml").read_text()
    file_path.write_text(
        aws_authmap_template.replace(
            "INSTANCEROLEARNHERE", node_instancerole_arn
        ).replace("ORIGINALPRINCIPALARNHERE", principal_arn)
    )
    return True


def remove(*, path: List[str], filename: str, **_: Any) -> bool:
    """Remove an EKS auth_map for worker connection.

    For use after destroying a cluster.

    Args:
        path: Path of the file.
        filename: Name of the file.

    Returns:
        boolean for whether or not the hook succeeded.

    """
    overlay_path = Path(*path)
    file_path = overlay_path / filename
    if file_path.is_file():
        LOGGER.info("Removing %s...", file_path)
        file_path.unlink()
    return True



"""CFNgin info action."""
import logging
from typing import Any

from .. import exceptions
from .base import BaseAction

LOGGER = logging.getLogger(__name__)


class Action(BaseAction):
    """Get information on CloudFormation stacks.

    Displays the outputs for the set of CloudFormation stacks.

    """

    NAME = "info"

    @property
    def _stack_action(self) -> Any:
        """Run against a step."""
        return None

    def run(self, *_args: Any, **_kwargs: Any) -> None:
        """Get information on CloudFormation stacks."""
        LOGGER.info("outputs for stacks: %s", self.context.get_fqn())
        if not self.context.stacks:
            LOGGER.warning("no stacks detected (error in config?)")
        for stack in self.context.stacks:
            provider = self.build_provider()

            try:
                provider_stack = provider.get_stack(stack.fqn)
            except exceptions.StackDoesNotExist:
                LOGGER.info("%s:stack does not exist", stack.fqn)
                continue

            LOGGER.info("%s:", stack.fqn)
            if "Outputs" in provider_stack:
                for output in provider_stack["Outputs"]:
                    LOGGER.info(
                        "\t%s: %s", output.get("OutputKey"), output.get("OutputValue")
                    )



"""Base classes for runway modules."""
from __future__ import annotations

import logging
import subprocess
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union, cast

from ..exceptions import NpmNotFound
from ..utils import which
from .utils import NPM_BIN, format_npm_command_for_logging, use_npm_ci

if TYPE_CHECKING:
    from .._logging import PrefixAdaptor, RunwayLogger
    from ..context import RunwayContext

LOGGER = cast("RunwayLogger", logging.getLogger(__name__))


class RunwayModule:
    """Base class for Runway modules."""

    ctx: RunwayContext
    explicitly_enabled: Optional[bool]
    logger: Union[PrefixAdaptor, RunwayLogger]
    name: str
    options: Union[Dict[str, Any], ModuleOptions]
    region: str

    def __init__(
        self,
        context: RunwayContext,
        *,
        explicitly_enabled: Optional[bool] = False,
        logger: RunwayLogger = LOGGER,
        module_root: Path,
        name: Optional[str] = None,
        options: Optional[Union[Dict[str, Any], ModuleOptions]] = None,
        parameters: Optional[Dict[str, Any]] = None,
        **_: Any,
    ) -> None:
        """Instantiate class.

        Args:
            context: Runway context object for the current session.
            explicitly_enabled: Whether or not the module is explicitly enabled.
                This is can be set in the event that the current environment being
                deployed to matches the defined environments of the module/deployment.
            logger: Used to write logs.
            module_root: Root path of the module.
            name: Name of the module.
            options: Options passed to the module class from the config as ``options``
                or ``module_options`` if coming from the deployment level.
            parameters: Values to pass to the underlying infrastructure as code
                tool that will alter the resulting infrastructure being deployed.
                Used to templatize IaC.

        """
        self.ctx = context
        self.explicitly_enabled = explicitly_enabled
        self.logger = logger
        self.name = name or module_root.name
        self.options = options or {}
        self.parameters = parameters or {}
        self.path = module_root
        self.region = context.env.aws_region

    def deploy(self) -> None:
        """Abstract method called when running deploy."""
        raise NotImplementedError("You must implement the deploy() method yourself!")

    def destroy(self) -> None:
        """Abstract method called when running destroy."""
        raise NotImplementedError("You must implement the destroy() method yourself!")

    def init(self) -> None:
        """Abstract method called when running init."""
        raise NotImplementedError("You must implement the init() method yourself!")

    def plan(self) -> None:
        """Abstract method called when running plan."""
        raise NotImplementedError("You must implement the plan() method yourself!")

    def __getitem__(self, key: str) -> Any:
        """Make the object subscriptable.

        Args:
            key: Attribute to get.

        """
        return getattr(self, key)


class RunwayModuleNpm(RunwayModule):  # pylint: disable=abstract-method
    """Base class for Runway modules that use npm."""

    def __init__(
        self,
        context: RunwayContext,
        *,
        explicitly_enabled: Optional[bool] = False,
        logger: RunwayLogger = LOGGER,
        module_root: Path,
        name: Optional[str] = None,
        options: Optional[Union[Dict[str, Any], ModuleOptions]] = None,
        parameters: Optional[Dict[str, Any]] = None,
        **_: Any,
    ) -> None:
        """Instantiate class.

        Args:
            context: Runway context object for the current session.
            explicitly_enabled: Whether or not the module is explicitly enabled.
                This is can be set in the event that the current environment being
                deployed to matches the defined environments of the module/deployment.
            logger: Used to write logs.
            module_root: Root path of the module.
            name: Name of the module.
            options: Options passed to the module class from the config as ``options``
                or ``module_options`` if coming from the deployment level.
            parameters: Values to pass to the underlying infrastructure as code
                tool that will alter the resulting infrastructure being deployed.
                Used to templatize IaC.

        """
        super().__init__(
            context,
            explicitly_enabled=explicitly_enabled,
            logger=logger,
            module_root=module_root,
            name=name,
            options=options,
            parameters=parameters,
        )
        self.check_for_npm(logger=self.logger)  # fail fast
        self.warn_on_boto_env_vars(self.ctx.env.vars, logger=logger)

    def log_npm_command(self, command: List[str]) -> None:
        """Log an npm command that is going to be run.

        Args:
            command: List that will be passed into a subprocess.

        """
        self.logger.debug("node command: %s", format_npm_command_for_logging(command))

    def npm_install(self) -> None:
        """Run ``npm install``."""
        if self.options.get("skip_npm_ci"):
            self.logger.info("skipped npm ci/npm install")
            return
        cmd = [NPM_BIN, "<place-holder>"]
        if self.ctx.no_color:
            cmd.append("--no-color")
        if self.ctx.is_noninteractive and use_npm_ci(self.path):
            self.logger.info("running npm ci...")
            cmd[1] = "ci"
        else:
            self.logger.info("running npm install...")
            cmd[1] = "install"
        subprocess.check_call(cmd)

    def package_json_missing(self) -> bool:
        """Check for the existence for a package.json file in the module.

        Returns:
            bool: True if the file was not found.

        """
        if not (self.path / "package.json").is_file():
            self.logger.debug("module is missing package.json")
            return True
        return False

    @staticmethod
    def check_for_npm(
        *, logger: Union[logging.Logger, PrefixAdaptor, RunwayLogger] = LOGGER
    ) -> None:
        """Ensure npm is installed and in the current path.

        Args:
            logger: Optionally provide a custom logger to use.

        """
        if not which("npm"):
            logger.error(
                '"npm" not found in path or is not executable; '
                "please ensure it is installed correctly"
            )
            raise NpmNotFound

    @staticmethod
    def warn_on_boto_env_vars(
        env_vars: Dict[str, str],
        *,
        logger: Union[logging.Logger, PrefixAdaptor, RunwayLogger] = LOGGER,
    ) -> None:
        """Inform user if boto-specific environment variables are in use.

        Args:
            env_vars: Environment variables to check.
            logger: Optionally provide a custom logger to use.

        """
        # https://github.com/serverless/serverless/issues/2151#issuecomment-255646512
        if env_vars.get("AWS_DEFAULT_PROFILE") and not env_vars.get("AWS_PROFILE"):
            logger.warning(
                "AWS_DEFAULT_PROFILE environment variable is set "
                "during use of nodejs-based module and AWS_PROFILE is "
                "not set -- you likely want to set AWS_PROFILE instead"
            )


class ModuleOptions:
    """Base class for Runway module options."""

    def get(self, name: str, default: Any = None) -> Any:
        """Get a value or return the default."""
        return getattr(self, name, default)

    def __eq__(self, other: Any) -> bool:
        """Assess equality."""
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return False



"""``runway run-python`` command."""
# docs: file://./../../../docs/source/commands.rst
import locale
from pathlib import Path
from typing import Any

import click

from .. import options


@click.command("run-python", short_help="bundled python")
@click.argument(
    "filename",
    metavar="<filename>",
    required=True,
    type=click.Path(exists=True, dir_okay=False),
)
@options.debug
@options.no_color
@options.verbose
def run_python(filename: str, **_: Any) -> None:
    """Execute a python script using a bundled copy of python.

    This command can execute actions using python without requiring python to
    be installed on a system.
    This is only applicable when installing a binary release of Runway
    (not installed from PyPi).
    When installed from PyPI, the current interpreter is used.

    """
    execglobals = globals().copy()
    # override name & file so script operates as if it were invoked directly
    execglobals.update({"__name__": "__main__", "__file__": filename})
    exec(  # pylint: disable=exec-used
        Path(filename).read_text(
            encoding=locale.getpreferredencoding(do_setlocale=False)
        ),
        execglobals,
        execglobals,
    )



"""CFNgin logger."""
import logging
import sys
from typing import Any, Dict, Optional

DEBUG_FORMAT = (
    "[%(asctime)s] %(levelname)s %(threadName)s "
    "%(name)s:%(lineno)d(%(funcName)s): %(message)s"
)
INFO_FORMAT = "[%(asctime)s] %(message)s"
COLOR_FORMAT = "[%(asctime)s] \033[%(color)sm%(message)s\033[39m"

ISO_8601 = "%Y-%m-%dT%H:%M:%S"


class ColorFormatter(logging.Formatter):
    """Handles colorizing formatted log messages if color provided."""

    def format(self, record: logging.LogRecord) -> str:
        """Format log message."""
        if "color" not in record.__dict__:
            record.__dict__["color"] = 37
        return super().format(record)


def setup_logging(verbosity: int, formats: Optional[Dict[str, Any]] = None):
    """Configure a proper logger based on verbosity and optional log formats.

    Args:
        verbosity: 0, 1, 2
        formats: Keys (``info``, ``color``, ``debug``) which may override the
            associated default log formats.

    """
    if formats is None:
        formats = {}

    log_level = logging.INFO

    log_format = formats.get("info", INFO_FORMAT)

    if sys.stdout.isatty():
        log_format = formats.get("color", COLOR_FORMAT)

    if verbosity > 0:
        log_level = logging.DEBUG
        log_format = formats.get("debug", DEBUG_FORMAT)

    if verbosity < 2:
        logging.getLogger("botocore").setLevel(logging.CRITICAL)

    hdlr = logging.StreamHandler()
    hdlr.setFormatter(ColorFormatter(log_format, ISO_8601))
    logging.root.addHandler(hdlr)
    logging.root.setLevel(log_level)



"""Shared jwt functionality for the Auth@Edge Lambda suite."""
import base64
import codecs
import json
import logging
import re
from urllib import request

from jose import jwt  # noqa pylint: disable=import-error

LOGGER = logging.getLogger(__name__)


def prepad_signed(hex_str):
    """Given a hexadecimal string prepad with 00 if not within range.

    Args:
        hex_str (string): The hexadecimal string.

    """
    msb = hex_str[0]
    if msb < "0" or msb > "7":
        return f"00{hex_str}"
    return hex_str


def to_hex(number):
    """Convert an integer to appropriate hex.

    Args:
        number (int): The number to convert.

    """
    n_str = format(int(number), "x")
    if len(n_str) % 2:
        return f"0{n_str}"
    return n_str


def encode_length_hex(number):
    """Encode the length value to a hexadecimal.

    Args:
        number (int): The number to convert.

    """
    if number <= 127:
        return to_hex(number)
    n_hex = to_hex(number)
    length = int(128 + len(n_hex) / 2)
    return to_hex(length) + n_hex


def rsa_public_key_to_pem(modulus_b64, exponent_b64):
    """Given an modulus and exponent convert an RSA public key to public PEM.

    Args:
        modulus_b64 (string): Base64 encoded modulus.
        exponent_b64 (string): Base64 encoded exponent.

    """
    modulus = base64.urlsafe_b64decode(modulus_b64 + "===")
    exponent = base64.urlsafe_b64decode(exponent_b64 + "===")
    modulus_hex = prepad_signed(modulus.hex())
    exponent_hex = prepad_signed(exponent.hex())
    mod_len = int(len(modulus_hex) / 2)
    exp_len = int(len(exponent_hex) / 2)

    encoded_mod_len = encode_length_hex(mod_len)
    encoded_exp_len = encode_length_hex(exp_len)

    encoded_pub_key = "30"
    encoded_pub_key += encode_length_hex(
        mod_len + exp_len + len(encoded_mod_len) / 2 + len(encoded_exp_len) / 2 + 2
    )
    encoded_pub_key += "02" + encoded_mod_len + modulus_hex
    encoded_pub_key += "02" + encoded_exp_len + exponent_hex

    der = base64.b64encode(codecs.decode(encoded_pub_key, "hex"))

    pem = "-----BEGIN RSA PUBLIC KEY-----\n"
    pem += "\n".join(re.findall(".{1,64}", der.decode()))
    pem += "\n-----END RSA PUBLIC KEY-----\n"
    return pem


class JwksClient:
    """Client responsible for retrieval of JWKS signing keys."""

    def __init__(self, options=None):
        """Initialize.

        Args:
            options (Optional[Dict[str, str]]): Options for the client.

        """
        self.options = options

    def get_keys(self):
        """Retrieve the keys from the JWKS endpoint."""
        LOGGER.info("Fetching keys from %s", self.options.get("jwks_uri"))

        try:
            # pylint: disable=consider-using-with
            request_res = request.urlopen(self.options.get("jwks_uri"))
            data = json.loads(
                request_res.read().decode(
                    request_res.info().get_param("charset") or "utf-8"
                )
            )
            keys = data["keys"]
            LOGGER.info("Keys: %s", keys)
            return keys
        except Exception as err:  # pylint: disable=broad-except
            LOGGER.info("Failure: ConnectionError")
            LOGGER.info(err)
            return {}

    def get_signing_key(self, kid):
        """Given a specific key id (kid) retrieve the signing key associated.

        Keyword Args:
            kid (str): The key id of the signing key.

        """
        LOGGER.info("Fetching signing key for %s", kid)

        keys = self.get_signing_keys()
        try:
            key = next(x for x in keys if x.get("kid") == kid)
            return key
        except StopIteration as exc:
            raise Exception(f"Was not able to locate a key with kid {kid}") from exc

    def get_signing_keys(self):
        """Given a set of keys find all that are signing keys."""
        keys = self.get_keys()

        if not keys:
            raise Exception("The JWKS endpoint did not contain any keys")

        jwks = []
        for key in filter(self.is_signing_key, keys):
            jwks.append(self.create_jwk(key))

        if not jwks:
            raise Exception("The JWKS endpoint did not contain any signing keys")

        LOGGER.info("Signing Keys: %s", jwks)
        return jwks

    @staticmethod
    def create_jwk(key):
        """Create the JSON Web Key.

        Keyword Args:
            key (dict): The Retrieved signing key.

        """
        jwk = {"kid": key.get("kid"), "nbf": key.get("nbf")}

        if key.get("x5c"):
            # @TODO: Support certificate chains. Review library here:
            # https://github.com/auth0/node-jwks-rsa/blob/master/src/JwksClient.js#L87
            LOGGER.info("X5C")
        else:
            try:
                jwk["rsaPublicKey"] = rsa_public_key_to_pem(key.get("n"), key.get("e"))
            # pylint: disable=broad-except
            except Exception as err:
                LOGGER.error(err)
                jwk["rsaPublicKey"] = None
        return jwk

    @staticmethod
    def is_signing_key(key):
        """Filter to determine if this is a signing key.

        Args:
            key (Dict[str, str]): The key.

        """
        if key.get("kty", "") != "RSA":
            return False
        if not key.get("kid", None):
            return False
        if key.get("use", "") != "sig":
            return False
        return key.get("x5c") or (key.get("n") and key.get("e"))


def is_rsa_signing_key(key):
    """Verify if the key specified is an RSA Public Key.

    Args:
        key (Dict): The key to filter.

    """
    return "rsaPublicKey" in key


def get_signing_key(jwks_uri, kid):
    """Retrieve the signing keys from the JWKS uri that match the key id specified.

    Args:
        jwks_uri (str): The URI in which to retrieve the JWKs.
        kid (str): Key ID of the signing key we are looking for.

    """
    client = JwksClient({"jwks_uri": jwks_uri})
    jwk = client.get_signing_key(kid)
    return jwk.get("rsaPublicKey") if is_rsa_signing_key(jwk) else jwk.get("publicKey")


def validate_jwt(jwt_token, jwks_uri, issuer, audience):
    """Validate the JWT token against the Cognito JWKs.

    Args:
        jwt_token (str): The JSON Web Token to validate.
        jwks_uri (str): The URI in which to retrieve the JSON Web Keys.
        issuer (str): Issuer of the JWT.
        audience (str): Audience of the JWT.

    """
    token_headers = jwt.get_unverified_header(jwt_token)
    if not token_headers:
        raise Exception("Cannot Parse JWT Token Headers")

    kid = token_headers.get("kid")
    jwk = get_signing_key(jwks_uri, kid)

    return jwt.decode(
        jwt_token,
        jwk,
        algorithms=["RS256"],
        audience=audience,
        issuer=issuer,
        options={"verify_at_hash": False},
    )


def validate_and_check_id_token(
    id_token, jwks_uri, issuer, audience, required_group=None
):
    """Validate JWT and (optionally) check group membership."""
    id_token_payload = validate_jwt(id_token, jwks_uri, issuer, audience)
    if required_group:
        cognito_groups = id_token_payload.get("cognito:groups")
        if not cognito_groups:
            raise MissingRequiredGroupError("Token does not have any groups")
        if required_group not in cognito_groups:
            raise MissingRequiredGroupError("Token does not have required group")


class MissingRequiredGroupError(Exception):
    """Raised when user is not in a required Cognito User Pool group."""



"""Add all configured (CloudFront compatable) headers to origin response."""
from shared import as_cloud_front_headers, get_config  # pylint: disable=import-error

CONFIG = get_config()


def handler(event, _context):
    """Handle adding the headers to the origin response.

    Args:
        event (Any): The Lambda Event.
        _context (Any): Lambda context object.

    """
    headers = CONFIG.get("http_headers")
    # Format to be CloudFront compatible
    configured_headers = as_cloud_front_headers(headers)
    response = event["Records"][0]["cf"]["response"]
    response["headers"].update(configured_headers)
    return response



"""Command hook."""
import logging
import os
import subprocess
from typing import Any, Dict, List, Optional, Union

from typing_extensions import TypedDict

from ...utils import BaseModel
from ..exceptions import ImproperlyConfigured

LOGGER = logging.getLogger(__name__)


class RunCommandHookArgs(BaseModel):
    """Hook arguments for ``run_command``."""

    capture: bool = False
    """If enabled, capture the command's stdout and stderr, and return them in the hook result."""

    command: Union[str, List[str]]
    """Command(s) to run."""

    env: Optional[Dict[str, str]] = None
    """Dictionary of environment variable overrides for the command context.
    Will be merged with the current environment.

    """

    ignore_status: bool = False
    """Don't fail the hook if the command returns a non-zero status."""

    interactive: bool = False
    """If enabled, allow the command to interact with stdin.
    Otherwise, stdin will be set to the null device.

    """

    quiet: bool = False
    """Redirect the command's stdout and stderr to the null device, silencing all output.
    Should not be enabled if ``capture`` is also enabled.

    """

    stdin: Optional[str] = None
    """String to send to the stdin of the command. Implicitly disables ``interactive``."""


class RunCommandResponseTypeDef(TypedDict, total=False):
    """Response from run_command."""

    returncode: int
    stderr: str
    stdout: str


def run_command(*__args: Any, **kwargs: Any) -> RunCommandResponseTypeDef:
    """Run a custom command as a hook.

    Arguments not parsed by the data model will be forwarded to the
    ``subprocess.Popen`` function. Interesting ones include: ``cwd`` and ``shell``.

    Examples:
        .. code-block:: yaml

            pre_deploy:
              command_copy_environment:
                path: runway.cfngin.hooks.command.run_command
                required: true
                enabled: true
                data_key: copy_env
                args:
                  command: ['cp', 'environment.template', 'environment']
              command_git_rev_parse:
                path: runway.cfngin.hooks.command.run_command
                required: true
                enabled: true
                data_key: get_git_commit
                args:
                  command: ['git', 'rev-parse', 'HEAD']
                  cwd: ./my-git-repo
                  capture: true
              command_npm_install:
                path: runway.cfngin.hooks.command.run_command
                args:
                  command: '`cd $PROJECT_DIR/project; npm install`'
                  env:
                    PROJECT_DIR: ./my-project
                    shell: true

    """
    args = RunCommandHookArgs.parse_obj(kwargs)

    # remove parsed args from kwargs
    for field in RunCommandHookArgs.__fields__:
        kwargs.pop(field, None)

    # remove unneeded args from kwargs
    kwargs.pop("context", None)
    kwargs.pop("provider", None)

    if args.quiet and args.capture:
        raise ImproperlyConfigured(
            __name__ + ".run_command",
            ValueError("Cannot enable `quiet` and `capture` options simultaneously"),
        )

    with open(os.devnull, "wb") as devnull:
        if args.quiet:
            out_err_type = devnull
        elif args.capture:
            out_err_type = subprocess.PIPE
        else:
            out_err_type = None

        if args.interactive:
            in_type = None
        elif args.stdin:
            in_type = subprocess.PIPE
        else:
            in_type = devnull

        if args.env:
            full_env = os.environ.copy()
            full_env.update(args.env)
            args.env = full_env

        LOGGER.info("running command: %s", args.command)

        with subprocess.Popen(
            args.command,
            stdin=in_type,
            stdout=out_err_type,
            stderr=out_err_type,
            env=args.env,
            **kwargs,
        ) as proc:
            try:
                out, err = proc.communicate(args.stdin)
                status = proc.wait()

                if status == 0 or args.ignore_status:
                    return {"returncode": proc.returncode, "stdout": out, "stderr": err}

                # Don't print the command line again if we already did earlier
                if LOGGER.isEnabledFor(logging.INFO):  # cov: ignore
                    LOGGER.warning("command failed with returncode %d", status)
                else:
                    LOGGER.warning(
                        "command failed with returncode %d: %s", status, args.command
                    )

                return {}
            except Exception:  # pylint: disable=broad-except  # cov: ignore
                return {}



"""Docker logic for the awslambda hook."""
from __future__ import annotations

import logging
import os
import platform
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Iterator,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
)

from docker import DockerClient
from docker.errors import DockerException, ImageNotFound
from docker.models.images import Image
from docker.types import Mount

from ...._logging import PrefixAdaptor
from ....compat import cached_property
from ....exceptions import DockerConnectionRefusedError, DockerExecFailedError
from .constants import AWS_SAM_BUILD_IMAGE_PREFIX, DEFAULT_IMAGE_NAME, DEFAULT_IMAGE_TAG

if TYPE_CHECKING:
    from pathlib import Path

    from ...._logging import RunwayLogger
    from ....context import CfnginContext, RunwayContext
    from .base_classes import Project
    from .models.args import AwsLambdaHookArgs, DockerOptions

LOGGER = cast("RunwayLogger", logging.getLogger(__name__))


_T = TypeVar("_T")


class DockerDependencyInstaller:
    """Docker dependency installer."""

    CACHE_DIR: ClassVar[str] = "/var/task/cache_dir"
    """Mount path where dependency managers can cache data."""

    DEPENDENCY_DIR: ClassVar[str] = "/var/task/lambda"
    """Mount path were dependencies will be installed to within the Docker container.
    Other files can be moved to this directory to be included in the deployment package.

    """

    PROJECT_DIR: ClassVar[str] = "/var/task/project"
    """Mount path where the project directory is available within the Docker container."""

    client: DockerClient
    """Docker client."""

    ctx: Union[CfnginContext, RunwayContext]
    """Context object."""

    options: DockerOptions
    """Hook arguments specific to Docker."""

    def __init__(
        self,
        project: Project[AwsLambdaHookArgs],
        *,
        client: Optional[DockerClient] = None,
        context: Optional[Union[CfnginContext, RunwayContext]] = None,
    ) -> None:
        """Instantiate class.

        This is a low-level method that requires the user to implement error
        handling. It is recommended to use
        :meth:`~runway.cfngin.hooks.awslambda.docker.DockerDependencyInstaller.from_project`
        instead of instantiating this class directly.

        Args:
            project: awslambda project.
            client: Pre-configured :class:`docker.client.DockerClient`.
            context: CFNgin or Runway context object.

        """
        context = context or project.ctx

        self._docker_logger = PrefixAdaptor("docker", LOGGER, "[{prefix}] {msg}")
        self.client = client or DockerClient.from_env(environment=context.env.vars)
        self.ctx = context
        self.options = project.args.docker
        self.project = project

    @cached_property
    def bind_mounts(self) -> List[Mount]:
        """Bind mounts that will be used by the container."""
        mounts = [
            Mount(
                target=self.DEPENDENCY_DIR,
                source=str(self.project.dependency_directory),
                type="bind",
            ),
            Mount(
                target=self.PROJECT_DIR,
                source=str(self.project.project_root),
                type="bind",
            ),
        ]
        if self.project.cache_dir:
            mounts.append(
                Mount(
                    target=self.CACHE_DIR,
                    source=str(self.project.cache_dir),
                    type="bind",
                )
            )
        return mounts

    @cached_property
    def environmet_variables(self) -> Dict[str, str]:
        """Environment variables to pass to the Docker container.

        This is a subset of the environment variables stored in the context
        object as some will cause issues if they are passed.

        """
        return {k: v for k, v in self.ctx.env.vars.items() if k.startswith("DOCKER")}

    @cached_property
    def image(self) -> Union[Image, str]:
        """Docker image that will be used.

        Raises:
            ValueError: Insufficient data to determine the desired Docker image.

        """
        if self.options.file:
            return self.build_image(self.options.file, name=self.options.name)
        if self.options.image:
            return self.pull_image(self.options.image, force=self.options.pull)
        if self.project.args.runtime:
            return self.pull_image(
                f"{AWS_SAM_BUILD_IMAGE_PREFIX}{self.project.args.runtime}:latest",
                force=self.options.pull,
            )
        raise ValueError("docker.file, docker.image, or runtime is required")

    @cached_property  # pylint error is python3.7 only
    def install_commands(self) -> List[str]:  # pylint: disable=no-self-use
        """Commands to run to install dependencies."""
        return []

    @cached_property
    def post_install_commands(self) -> List[str]:
        """Commands to run after dependencies have been installed."""
        cmds = [
            *[
                # wildcards need to exist outside of the quotes to work
                # needs to be wrapped in `sh -c` to resolve wildcard
                f"sh -c 'cp -v \"{extra_file.rstrip('*')}\"* \"{self.DEPENDENCY_DIR}\"'"
                if extra_file.endswith("*")
                else f'sh -c \'cp -v "{extra_file}" "{self.DEPENDENCY_DIR}"\''
                for extra_file in self.options.extra_files
            ],
        ]
        if platform.system() != "Windows":
            # methods only exist on POSIX systems
            gid, uid = os.getgid(), os.getuid()  # pylint: disable=no-member
            cmds.append(
                f"chown -R {uid}:{gid} {self.DEPENDENCY_DIR}",
            )
            if self.project.cache_dir:
                cmds.append(f"chown -R {uid}:{gid} {self.CACHE_DIR}")
        return cmds

    @cached_property
    def pre_install_commands(self) -> List[str]:
        """Commands to run before dependencies have been installed."""
        cmds = [
            f"chown -R 0:0 {self.DEPENDENCY_DIR}",
        ]
        if self.project.cache_dir:
            cmds.append(f"chown -R 0:0 {self.CACHE_DIR}")
        return cmds

    @cached_property  # pylint error is python3.7 only
    def runtime(self) -> Optional[str]:  # pylint: disable=no-self-use
        """AWS Lambda runtime determined from the Docker container."""
        return None

    def build_image(
        self,
        docker_file: Path,
        *,
        name: Optional[str] = None,
        tag: Optional[str] = None,
    ) -> Image:
        """Build Docker image from Dockerfile.

        This method is exposed as a low-level interface.
        :attr:`~runway.cfngin.hooks.awslambda.docker.DockerDependencyInstaller.image`
        should be used in place for this for most cases.

        Args:
            docker_file: Path to the Dockerfile to build. This path should be
                absolute, must exist, and must be a file.
            name: Name of the Docker image. The name should not contain a tag.
                If not provided, a default value is use.
            tag: Tag to apply to the image after it is built.
                If not provided, a default value of ``latest`` is used.

        Returns:
            Object representing the image that was built.

        """
        image, log_stream = self.client.images.build(
            dockerfile=docker_file.name,
            forcerm=True,
            path=str(docker_file.parent),
            pull=self.options.pull,
        )
        self.log_docker_msg_dict(log_stream)
        image.tag(name or DEFAULT_IMAGE_NAME, tag=tag or DEFAULT_IMAGE_TAG)
        image.reload()
        LOGGER.info("built docker image %s (%s)", ", ".join(image.tags), image.id)
        return image

    def log_docker_msg_bytes(
        self, stream: Iterator[bytes], *, level: int = logging.INFO
    ) -> List[str]:
        """Log Docker output message from blocking generator that return bytes.

        Args:
            stream: Blocking generator that returns log messages as bytes.
            level: Log level to use when logging messages.

        Returns:
            List of log messages.

        """
        result: List[str] = []
        for raw_msg in stream:
            msg = raw_msg.decode().strip()
            result.append(msg)
            self._docker_logger.log(level, msg)
        return result

    def log_docker_msg_dict(
        self, stream: Iterator[Dict[str, Any]], *, level: int = logging.INFO
    ) -> List[str]:
        """Log Docker output message from blocking generator that return dict.

        Args:
            stream: Blocking generator that returns log messages as a dict.
            level: Log level to use when logging messages.

        Returns:
            list of log messages.

        """
        result: List[str] = []
        for raw_msg in stream:
            for key in ["stream", "status"]:
                if key in raw_msg:
                    msg = raw_msg[key].strip()
                    result.append(msg)
                    self._docker_logger.log(level, msg)
                    break
        return result

    def install(self) -> None:
        """Install dependencies using Docker.

        Commands are run as they are defined in the following cached properties:

        - :attr:`~runway.cfngin.hooks.awslambda.docker.DockerDependencyInstaller.pre_install_commands`
        - :attr:`~runway.cfngin.hooks.awslambda.docker.DockerDependencyInstaller.install_commands`
        - :attr:`~runway.cfngin.hooks.awslambda.docker.DockerDependencyInstaller.post_install_commands`

        """  # noqa
        for cmd in self.pre_install_commands:
            self.run_command(cmd)
        for cmd in self.install_commands:
            self.run_command(cmd)
        for cmd in self.post_install_commands:
            self.run_command(cmd)

    def pull_image(self, name: str, *, force: bool = True) -> Image:
        """Pull a Docker image from a repository if it does not exist locally.

        This method is exposed as a low-level interface.
        :attr:`~runway.cfngin.hooks.awslambda.docker.DockerDependencyInstaller.image`
        should be used in place for this for most cases.

        Args:
            name: Name of the Docker image including tag.
            force: Always pull the image even if it exists locally.
                This will ensure that the latest version is always used.

        Returns:
            Object representing the image found locally or pulled from a repository.

        """
        try:
            if not force:
                return self.client.images.get(name)
            LOGGER.info("pulling docker image %s...", name)
        except ImageNotFound:
            LOGGER.info("image not found; pulling docker image %s...", name)
        return self.client.images.pull(repository=name)

    def run_command(self, command: str, *, level: int = logging.INFO) -> List[str]:
        """Execute equivalent of ``docker container run``.

        Args:
            command: Command to be run.
            level: Log level to use when logging messages.

        Raises:
            DockerExecFailedError: Docker container returned a non-zero exit code.

        Returns:
            List of log messages.

        """
        LOGGER.verbose("running command with docker: %s", command)
        container = self.client.containers.create(
            command=command,
            detach=True,
            environment=self.environmet_variables,
            image=self.image,
            mounts=self.bind_mounts,
            working_dir=self.PROJECT_DIR,
        )
        try:
            container.start()
            return self.log_docker_msg_bytes(
                container.logs(stderr=True, stdout=True, stream=True), level=level
            )
        finally:
            response = container.wait()
            container.remove(force=True)  # always remove container
            if response.get("StatusCode", 0) != 0:
                raise DockerExecFailedError(response)

    @classmethod
    def from_project(
        cls: Type[_T], project: Project[AwsLambdaHookArgs]
    ) -> Optional[_T]:
        """Instantiate class from a project.

        High-level method that wraps instantiation in error handling.

        Args:
            project: Project being processed.

        Returns:
            Object to handle dependency installation with Docker if Docker is
            available and not disabled.

        Raises:
            DockerConnectionRefused: Docker is not install or is unreachable.

        """
        if project.args.docker.disabled:
            return None
        try:
            client = DockerClient.from_env(environment=project.ctx.env.vars)
            if client.ping():
                return cls(project, client=client)
        except DockerException as exc:
            # This might be too broad but, it is the only repeated substring
            # between operating systems in similar states.
            if "Error while fetching server API version" in str(exc):
                raise DockerConnectionRefusedError from exc
            raise
        # ping failed but method did not return false for some reason
        raise DockerConnectionRefusedError



"""Run a kubectl command."""
# docs: file://./../../../../docs/source/commands.rst
import logging
import subprocess
from typing import Any, Tuple

import click

from ....env_mgr.kbenv import KBEnvManager
from ... import options

LOGGER = logging.getLogger(__name__.replace("._", "."))


@click.command(
    "run", short_help="run kubectl", context_settings={"ignore_unknown_options": True}
)
@click.argument("args", metavar="<args>", nargs=-1, required=True)
@options.debug
@options.no_color
@options.verbose
@click.pass_context
def run(ctx: click.Context, args: Tuple[str, ...], **_: Any) -> None:
    """Run a kubectl command.

    Uses the version of kubectl specified in the ".kubectl-version" file
    in the current directory.

    IMPORTANT: When using options shared with Runway "--" must be placed
    before the kubectl command.

    """
    ctx.exit(subprocess.call([KBEnvManager().install()] + list(args)))



"""``runway kbenv`` command group."""
# docs: file://./../../../../docs/source/commands.rst
from typing import Any, List

import click

from ... import options
from ._install import install
from ._list import list_installed
from ._run import run
from ._uninstall import uninstall

__all__ = ["install", "list_installed", "run", "uninstall"]

COMMANDS: List[click.Command] = [install, list_installed, run, uninstall]


@click.group("kbenv", short_help="kubectl (install|run)")
@options.debug
@options.no_color
@options.verbose
def kbenv(**_: Any) -> None:
    """kubectl version management and execution.

    Compatible with https://github.com/alexppg/kbenv.

    """


for cmd in COMMANDS:  # register commands
    kbenv.add_command(cmd)



"""Botocore with support for AWS SSO utilities."""
# Copyright 2012-2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import hashlib
import logging

from .exceptions import SSOTokenLoadError

LOGGER = logging.getLogger(__name__)


class SSOTokenLoader:
    """AWS SSO token loader."""

    def __init__(self, cache=None):
        """Instantiate class."""
        if cache is None:
            cache = {}
        self._cache = cache

    @staticmethod
    def _generate_cache_key(start_url):
        """Generate cache key."""
        return hashlib.sha1(start_url.encode("utf-8")).hexdigest()

    def __call__(self, start_url):
        """Call instance of class directly."""
        cache_key = self._generate_cache_key(start_url)
        try:
            token = self._cache[cache_key]
            return token["accessToken"]
        except KeyError as exc:
            LOGGER.debug("Failed to load SSO token:", exc_info=True)
            error_msg = (
                "The SSO access token has either expired or is otherwise invalid."
            )
            raise SSOTokenLoadError(error_msg=error_msg) from exc



"""AWS type definitions."""
from __future__ import annotations

from typing import List

from typing_extensions import TypedDict


class TagTypeDef(TypedDict):
    """AWS resource tags."""

    Key: str
    Value: str


TagSetTypeDef = List[TagTypeDef]



"""Serverless Hello World function."""
from __future__ import annotations

import json
from typing import Any, Dict, Union


# pylint: disable=unused-argument
def handler(event: Any, context: Any) -> Dict[str, Union[int, str]]:
    """Return Serverless Hello World."""
    body = {
        "message": "Go Serverless v1.0! Your function executed successfully!",
        "input": event,
    }

    response = {"statusCode": 200, "body": json.dumps(body)}

    return response

    # Use this code if you don't use the http event with the LAMBDA-PROXY
    # integration
    # return {
    #     "message": "Go Serverless v1.0! Your function executed!",
    #     "event": event
    # }



"""Runway CLI entrypoint."""
import argparse
import logging
import os
from typing import Any, Dict

import click

from runway import __version__

from . import commands, options
from .logs import setup_logging
from .utils import CliContext

LOGGER = logging.getLogger("runway.cli")

CLICK_CONTEXT_SETTINGS: Dict[str, Any] = {
    "help_option_names": ["-h", "--help"],
    "max_content_width": 999,
}


class _CliGroup(click.Group):
    """Extends the use of click.Group.

    This should only be used for the main application group.

    """

    def invoke(self, ctx: click.Context) -> Any:
        """Replace invoke command to pass along args."""
        ctx.meta["global.options"] = self.__parse_global_options(ctx)
        return super().invoke(ctx)

    @staticmethod
    def __parse_global_options(ctx: click.Context) -> Dict[str, Any]:
        """Parse global options.

        These options are passed to subcommands but, should be parsed by the
        main application group. The value of these options are used for global
        configuration such as logging or context object setup.

        """
        parser = argparse.ArgumentParser(add_help=False)
        parser.add_argument("--ci", action="store_true", default=bool(os.getenv("CI")))
        parser.add_argument(
            "--debug", default=int(os.getenv("DEBUG", "0")), action="count"
        )
        parser.add_argument(
            "-e", "--deploy-environment", default=os.getenv("DEPLOY_ENVIRONMENT")
        )
        parser.add_argument(
            "--no-color",
            action="store_true",
            default=bool(os.getenv("RUNWAY_NO_COLOR")),
        )
        parser.add_argument(
            "--verbose", action="store_true", default=bool(os.getenv("VERBOSE"))
        )
        args, _ = parser.parse_known_args(list(ctx.args))
        return vars(args)


@click.group(context_settings=CLICK_CONTEXT_SETTINGS, cls=_CliGroup)
@click.version_option(__version__, message="%(version)s")
@options.debug
@options.no_color
@options.verbose
@click.pass_context
def cli(ctx: click.Context, **_: Any) -> None:
    """Runway CLI.

    Full documentation available at https://docs.onica.com/projects/runway/

    """
    opts = ctx.meta["global.options"]
    setup_logging(
        debug=opts["debug"], no_color=opts["no_color"], verbose=opts["verbose"]
    )
    ctx.obj = CliContext(**opts)


# register all the other commands from the importable modules defined
# in commands.
for cmd in commands.__all__:
    cli.add_command(getattr(commands, cmd))



"""Base module for environment managers."""
from __future__ import annotations

import logging
import os
import platform
import shutil
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Generator, Optional, Union, cast

from ..compat import cached_property
from ..mixins import DelCachedPropMixin

if TYPE_CHECKING:
    from urllib.error import URLError

    from .._logging import RunwayLogger
    from ..utils import Version

LOGGER = cast("RunwayLogger", logging.getLogger(__name__))


def handle_bin_download_error(exc: URLError, name: str) -> None:
    """Give user info about their failed download.

    Raises:
        SystemExit: Always raised after logging reason.

    """
    url_error_msg = str(exc.reason)

    if "CERTIFICATE_VERIFY_FAILED" not in url_error_msg:
        raise exc

    LOGGER.error(
        "Attempted to download %s but was unable to verify the TLS "
        "certificate on its download site.",
        name,
    )
    LOGGER.error("Full TLS error message: %s", url_error_msg)
    if platform.system().startswith("Darwin") and (
        "unable to get local issuer certificate" in url_error_msg
    ):
        LOGGER.error(
            "This is likely caused by your Python installation missing root certificates. "
            'Run "/Applications/Python %s.%s/"Install Certificates.command" to fix it '
            "(https://stackoverflow.com/a/42334357/2547802)",
            sys.version_info[0],
            sys.version_info[1],
        )
    sys.exit(1)


class EnvManager(DelCachedPropMixin):
    """Base environment manager class.

    Attributes:
        binPath to the binary of the current version.
        current_version: The current binary version being used.
        env_dir_name: Name of the directory within the users home
            directory where binary versions will be stored.
        path: The current working directory.

    """

    _bin_name: str

    current_version: Optional[str]
    env_dir_name: str
    path: Path

    def __init__(
        self, bin_name: str, dir_name: str, path: Optional[Path] = None
    ) -> None:
        """Initialize class.

        Args:
            bin_name: Name of the binary file (e.g. kubectl)
            dir_name: Name of the directory within the users home
                directory where binary versions will be stored.
            path: The current working directory.

        """
        self._bin_name = bin_name + self.command_suffix
        self.current_version = None
        self.env_dir_name = (
            dir_name if platform.system() == "Windows" else "." + dir_name
        )
        self.path = Path.cwd() if not path else path

    @property
    def bin(self) -> Path:
        """Path to the version binary.

        Returns:
            Path

        """
        if self.current_version:
            return self.versions_dir / self.current_version / self._bin_name
        return self.versions_dir / self._bin_name

    @cached_property
    def command_suffix(self) -> str:  # pylint: disable=no-self-use
        """Return command suffix based on platform.system."""
        if platform.system() == "Windows":
            return ".exe"
        return ""

    @cached_property
    def env_dir(self) -> Path:
        """Return the directory used to store version binaries."""
        if platform.system() == "Windows":
            if "APPDATA" in os.environ:
                return Path(os.environ["APPDATA"]) / self.env_dir_name
            return Path.home() / "AppData" / "Roaming" / self.env_dir_name
        return Path.home() / self.env_dir_name

    @cached_property
    def versions_dir(self) -> Path:
        """Return the directory used to store binary.

        When first used, the existence of the directory is checked and it is
        created if needed.

        """
        return self.env_dir / "versions"

    @cached_property
    def version_file(self) -> Optional[Path]:
        """Find and return a "<bin version file>" file if one is present.

        Returns:
            Path to the <bin> version file.

        """
        raise NotImplementedError

    def install(self, version_requested: Optional[str] = None) -> str:
        """Ensure <bin> is installed."""
        raise NotImplementedError

    def list_installed(self) -> Generator[Path, None, None]:
        """List installed versions of <bin>."""
        raise NotImplementedError

    def uninstall(self, version: Union[str, Version]) -> bool:
        """Uninstall a version of the managed binary.

        Args:
            version: Version of binary to uninstall.

        Returns:
            Whether a version of the binary was uninstalled or not.

        """
        version_dir = self.versions_dir / str(version)
        if version_dir.is_dir():
            LOGGER.notice(
                "uninstalling %s %s from %s...",
                self._bin_name,
                version,
                self.versions_dir,
            )
            shutil.rmtree(version_dir)
            LOGGER.success("uninstalled %s %s", self._bin_name, version)
            return True
        LOGGER.error("%s %s not installed", self._bin_name, version)
        return False



"""Docker login hook."""
# pylint: disable=no-self-argument,no-self-use
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Dict, Optional

from pydantic import Field, validator

from ....utils import BaseModel
from .data_models import ElasticContainerRegistry
from .hook_data import DockerHookData

if TYPE_CHECKING:
    from ....context import CfnginContext

LOGGER = logging.getLogger(__name__.replace("._", "."))


class LoginArgs(BaseModel):
    """Args passed to the docker.login hook."""

    _ctx: Optional[CfnginContext] = Field(default=None, alias="context", exclude=True)

    dockercfg_path: Optional[str] = None
    """Path to a non-default Docker config file."""

    ecr: Optional[ElasticContainerRegistry] = Field(default=None, exclude=True)
    """Information describing an ECR registry."""

    email: Optional[str] = None
    """The email for the registry account."""

    password: str
    """The plaintext password for the registry account."""

    registry: Optional[str] = None
    """URI of the registry to login to."""

    username: str = "AWS"
    """The registry username."""

    @validator("ecr", pre=True, allow_reuse=True)
    def _set_ecr(cls, v: Any, values: Dict[str, Any]) -> Any:
        """Set the value of ``ecr``."""
        if v and isinstance(v, dict):
            return ElasticContainerRegistry.parse_obj(
                {"context": values.get("context"), **v}
            )
        return v

    @validator("registry", pre=True, always=True, allow_reuse=True)
    def _set_registry(cls, v: Any, values: Dict[str, Any]) -> Any:
        """Set the value of ``registry``."""
        if v:
            return v

        ecr: Optional[ElasticContainerRegistry] = values.get("ecr")
        if ecr:
            return ecr.fqn

        return None


def login(*, context: CfnginContext, **kwargs: Any) -> DockerHookData:
    """Docker login hook.

    Replicates the functionality of ``docker login`` cli command.

    kwargs are parsed by :class:`~runway.cfngin.hooks.docker.LoginArgs`.

    """
    args = LoginArgs.parse_obj({"context": context, **kwargs})
    docker_hook_data = DockerHookData.from_cfngin_context(context)
    docker_hook_data.client.login(**args.dict())
    LOGGER.info("logged into %s", args.registry)
    return docker_hook_data.update_context(context)



"""``runway gen-sample tf`` command."""
import locale
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import click

from ....env_mgr.tfenv import get_latest_tf_version
from ... import options
from .utils import TEMPLATES, copy_sample

if TYPE_CHECKING:
    from ...._logging import RunwayLogger

LOGGER = cast("RunwayLogger", logging.getLogger(__name__.replace("._", ".")))


@click.command("tf", short_help="tf (sampleapp.tf)")
@options.debug
@options.no_color
@options.verbose
@click.pass_context
def tf(ctx: click.Context, **_: Any) -> None:  # pylint: disable=invalid-name
    """Generate a sample Terraform project."""
    src = TEMPLATES / "terraform"
    dest = Path.cwd() / "sampleapp.tf"

    copy_sample(ctx, src, dest)

    if not (src / ".terraform-version").is_file():
        (dest / ".terraform-version").write_text(
            get_latest_tf_version(),
            encoding=locale.getpreferredencoding(do_setlocale=False),
        )

    LOGGER.success("Sample Terraform app created at %s", dest)



"""Runway AWS Cloud Development Kit Module options."""
from __future__ import annotations

from typing import List

from pydantic import Extra

from ...base import ConfigProperty


class RunwayCdkModuleOptionsDataModel(ConfigProperty):
    """Model for Runway AWS Cloud Development Kit Module options."""

    build_steps: List[str] = []
    skip_npm_ci: bool = False

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.ignore
        title = "Runway AWS Cloud Development Kit Module options."



"""``runway destroy`` command."""
# docs: file://./../../../docs/source/commands.rst
import logging
from typing import Any, Tuple

import click
from pydantic import ValidationError

from ...core import Runway
from ...exceptions import ConfigNotFound, VariablesFileNotFound
from .. import options
from ..utils import select_deployments

LOGGER = logging.getLogger(__name__.replace("._", "."))


@click.command("destroy", short_help="destroy things")
@options.ci
@options.debug
@options.deploy_environment
@options.no_color
@options.tags
@options.verbose
@click.pass_context
def destroy(ctx: click.Context, debug: bool, tags: Tuple[str, ...], **_: Any) -> None:
    """Destroy infrastructure as code.

    \b
    Process
    -------
    1. Determines the deploy environment.
        - "-e, --deploy-environment" option
        - "DEPLOY_ENVIRONMENT" environment variable
        - git branch name
            - strips "ENV-" prefix, master is converted to common
            - ignored if "ignore_git_branch: true"
        - name of the current working directory
    2. Selects deployments & modules to deploy.
        - (default) prompts
        - (tags) module contains all tags
        - (non-interactive) all
    3. Destroys selected deployments/modules in reverse the order defined.

    """  # noqa: D301
    if not ctx.obj.env.ci:
        click.secho(
            "[WARNING] Runway is about to be run in DESTROY mode. [WARNING]",
            bold=True,
            fg="red",
        )
        click.secho(
            "Any/all deployment(s) selected will be irrecoverably DESTROYED.",
            bold=True,
            fg="red",
        )
        if not click.confirm("\nProceed?"):
            ctx.exit(0)
        click.echo("")
    try:
        Runway(ctx.obj.runway_config, ctx.obj.get_runway_context()).destroy(
            Runway.reverse_deployments(
                select_deployments(ctx, ctx.obj.runway_config.deployments, tags)
            )
        )
    except ValidationError as err:
        LOGGER.error(err, exc_info=debug)
        ctx.exit(1)
    except (ConfigNotFound, VariablesFileNotFound) as err:
        LOGGER.error(err.message, exc_info=debug)
        ctx.exit(1)



"""Run a kubectl command."""
# docs: file://./../../../../docs/source/commands.rst
import logging
import subprocess
from typing import Any, Tuple

import click

from ....env_mgr.tfenv import TFEnvManager
from ....utils import DOC_SITE
from ... import options

LOGGER = logging.getLogger(__name__.replace("._", "."))


@click.command(
    "run", short_help="run terraform", context_settings={"ignore_unknown_options": True}
)
@click.argument("args", metavar="<args>", nargs=-1, required=True)
@options.debug
@options.no_color
@options.verbose
@click.pass_context
def run(ctx: click.Context, args: Tuple[str, ...], **_: Any) -> None:
    """Run a Terraform command.

    Uses the version of Terraform specified in the ".terraform-version" file
    in the current directory.

    IMPORTANT: When using options shared with Runway "--" must be placed
    before the Terraform command.

    """
    try:
        ctx.exit(subprocess.call([TFEnvManager().install()] + list(args)))
    except ValueError as err:
        LOGGER.debug("terraform install failed", exc_info=True)
        if "unable to find" not in str(err):
            LOGGER.error(
                "unexpected error encountered when trying to install Terraform",
                exc_info=True,
            )
            ctx.exit(1)
        else:
            LOGGER.error("unable to find a .terraform-version file")
            LOGGER.error(
                "learn how to use Runway to manage Terraform versions at "
                "%s/page/terraform/advanced_features.html#version-management",
                DOC_SITE,
            )
        ctx.exit(1)



"""CFNgin hooks for AWS Certificate Manager."""
from __future__ import annotations

import logging
import time
from typing import TYPE_CHECKING, Any, ClassVar, Dict, List, Optional, Type

from botocore.exceptions import ClientError
from troposphere import Ref
from troposphere.certificatemanager import Certificate as CertificateResource
from typing_extensions import Literal

from ...utils import MutableMap
from ..blueprints.variables.types import CFNString
from ..exceptions import StackDoesNotExist, StackFailed, StackUpdateBadStatus
from ..status import NO_CHANGE, SUBMITTED
from .base import Hook, HookArgsBaseModel
from .utils import BlankBlueprint

if TYPE_CHECKING:
    from mypy_boto3_acm.client import ACMClient
    from mypy_boto3_acm.type_defs import ResourceRecordTypeDef
    from mypy_boto3_route53.client import Route53Client
    from mypy_boto3_route53.type_defs import ChangeTypeDef

    from ...context import CfnginContext
    from ..blueprints.base import Blueprint
    from ..providers.aws.default import Provider
    from ..stack import Stack
    from ..status import Status

LOGGER = logging.getLogger(__name__)


class HookArgs(HookArgsBaseModel):
    """Hook arguments."""

    alt_names: List[str] = []
    domain: str
    hosted_zone_id: str
    stack_name: Optional[str] = None
    ttl: int = 300


class Certificate(Hook):
    r"""Hook for managing a **AWS::CertificateManager::Certificate**.

    Keyword Args:
        alt_names (Optional[List[str]]): Additional FQDNs to be included in the
            Subject Alternative Name extension of the ACM certificate. For
            example, you can add www.example.net to a certificate for which the
            domain field is www.example.com if users can reach your site by
            using either name.
        domain (str): The fully qualified domain name (FQDN), such as
            www.example.com, with which you want to secure an ACM certificate.
            Use an asterisk (``*``) to create a wildcard certificate that
            protects several sites in the same domain. For example,
            \*.example.com protects www.example.com, site.example.com, and
            images.example.com.
        hosted_zone_id (str): The ID of the Route 53 Hosted Zone that contains
            the resource record sets that you want to change. This must exist
            in the same account that the certificate will be created in.
        stack_name (Optional[str]): Provide a name for the stack used to
            create the certificate. If not provided, the domain is used
            (replacing ``.`` with ``-``).
        ttl (Optional[int]): The resource record cache time to live (TTL),
            in seconds. (*default:* ``300``)

    .. rubric:: Example
    .. code-block:: yaml

        pre_deploy:
          example-wildcard-cert:
            path: runway.cfngin.hooks.acm.Certificate
            required: true
            args:
              domain: '*.example.com'
              hosted_zone_id: ${xref example-com::HostedZoneId}

    """

    ARGS_PARSER: ClassVar[Type[HookArgs]] = HookArgs

    acm_client: ACMClient
    args: HookArgs
    blueprint: Blueprint
    r53_client: Route53Client
    stack: Stack
    template_description: str

    def __init__(
        self, context: CfnginContext, provider: Provider, **kwargs: Any
    ) -> None:
        """Instantiate class.

        Args:
            context: Context instance. (passed in by CFNgin)
            provider: Provider instance. (passed in by CFNgin)

        """
        super().__init__(context, provider, **kwargs)

        self.template_description = self.get_template_description()
        self.stack_name = self.args.stack_name or self.args.domain.replace(".", "-")

        self.properties = MutableMap(
            **{
                "DomainName": self.args.domain,
                "SubjectAlternativeNames": self.args.alt_names,
                "Tags": self.tags,
                "ValidationMethod": "DNS",
            }
        )
        self.blueprint = self._create_blueprint()

        session = self.context.get_session()
        self.acm_client = session.client("acm")
        self.r53_client = session.client("route53")
        self.stack = self.generate_stack(
            variables={
                "ValidateRecordTTL": self.args.ttl,
                "DomainName": self.args.domain,
            }
        )

    def _create_blueprint(self) -> Blueprint:
        """Create CFNgin Blueprint."""
        var_description = (
            "NO NOT CHANGE MANUALLY! Used to track the "
            "state of a value set outside of CloudFormation"
            " using a Runway hook."
        )

        class _BlankBlueprint(BlankBlueprint):
            VARIABLES = {
                "DomainName": {"type": CFNString, "description": var_description},
                "ValidateRecordTTL": {
                    "type": CFNString,
                    "description": var_description,
                },
            }

        blueprint = _BlankBlueprint(self.stack_name, self.context)
        blueprint.template.set_version("2010-09-09")
        blueprint.template.set_description(self.template_description)

        cert = blueprint.template.add_resource(
            CertificateResource("Certificate", **self.properties.data)
        )
        blueprint.add_output(f"{cert.title}Arn", cert.ref())
        blueprint.add_output("ValidateRecordTTL", Ref("ValidateRecordTTL"))
        blueprint.add_output("DomainName", Ref("DomainName"))
        return blueprint

    def domain_changed(self) -> bool:
        """Check to ensure domain has not changed for existing stack."""
        try:
            stack_info = self.provider.get_stack(self.stack.fqn)
            if self.provider.is_stack_recreatable(stack_info):
                LOGGER.debug(
                    "stack is in a recreatable state; domain change does not matter"
                )
                return False
            if self.provider.is_stack_in_progress(
                stack_info
            ) or self.provider.is_stack_rolling_back(stack_info):
                LOGGER.debug("stack is in progress; can't check for domain change")
                return False
            if (
                self.args.domain
                != self.provider.get_outputs(self.stack.fqn)["DomainName"]
            ):
                LOGGER.error(
                    '"domain" can\'t be changed for existing '
                    'certificate in stack "%s"',
                    self.stack.fqn,
                )
                return True
        except StackDoesNotExist:
            pass
        except KeyError:
            LOGGER.warning(
                'stack "%s" is missing output DomainName; update may fail',
                self.stack.fqn,
            )
        return False

    def get_certificate(self, interval: int = 5) -> str:
        """Get the certificate being created by a CloudFormation.

        Args:
            interval: Number of seconds to wait between attempts.

        Returns:
            Certificate ARN.

        """
        response = self.provider.cloudformation.describe_stack_resources(
            StackName=self.stack.fqn, LogicalResourceId="Certificate"
        )["StackResources"]
        if response:
            # can be returned without having a PhysicalResourceId
            if "PhysicalResourceId" in response[0]:
                return response[0]["PhysicalResourceId"]
        LOGGER.debug("waiting for certificate to be created...")
        time.sleep(interval)
        return self.get_certificate(interval=interval)

    def get_validation_record(
        self,
        cert_arn: Optional[str] = None,
        *,
        interval: int = 5,
        status: str = "PENDING_VALIDATION",
    ) -> ResourceRecordTypeDef:
        """Get validation record from the certificate being created.

        Args:
            cert_arn: ARN of the certificate to validate.
            interval: Number of seconds to wait between attempts.
            status: Validation status to look for when finding a validation record.
                Typically only "PENDING_VALIDATION" or "SUCCESS" will be used.

        Raises:
            ValueError: No pending or too many pending certificates.

        """
        if not cert_arn:
            cert_arn = self.get_certificate()
        cert = self.acm_client.describe_certificate(CertificateArn=cert_arn).get(
            "Certificate", {}
        )

        try:
            domain_validation = [
                opt
                for opt in cert["DomainValidationOptions"]
                if opt["ValidationStatus"] == status
            ]
        except KeyError:
            LOGGER.debug(
                "waiting for DomainValidationOptions to become "
                "available for the certificate..."
            )
            time.sleep(interval)
            return self.get_validation_record(
                cert_arn=cert_arn, interval=interval, status=status
            )

        if not domain_validation:
            raise ValueError(
                f'No validations with status "{status}" found for "{self.args.domain}"'
            )
        if len(domain_validation) > 1:
            raise ValueError(
                f"Found {len(domain_validation)} validation options of status "
                f'"{status}" for "{self.args.domain}"; only one option is supported'
            )
        try:
            # the validation option can exists before the record set is ready
            return domain_validation[0]["ResourceRecord"]
        except KeyError:
            LOGGER.debug(
                "waiting for DomainValidationOptions.ResourceRecord "
                "to become available for the certificate..."
            )
            time.sleep(interval)
            return self.get_validation_record(
                cert_arn=cert_arn, interval=interval, status=status
            )

    def put_record_set(self, record_set: ResourceRecordTypeDef) -> None:
        """Create/update a record set on a Route 53 Hosted Zone.

        Args:
            record_set: Record set to be added to Route 53.

        """
        LOGGER.info(
            "adding validation record to hosted zone: %s", self.args.hosted_zone_id
        )
        self.__change_record_set("CREATE", [record_set])

    def remove_validation_records(
        self, records: Optional[List[ResourceRecordTypeDef]] = None
    ) -> None:
        """Remove all record set entries used to validate an ACM Certificate.

        Args:
            records: List of validation records to remove from Route 53.
                This can be provided in cases were the certificate has been
                deleted during a rollback.

        """
        if not records:
            cert_arn = self.get_certificate()
            cert = self.acm_client.describe_certificate(CertificateArn=cert_arn).get(
                "Certificate", {}
            )

            records = [
                opt["ResourceRecord"]
                for opt in cert.get("DomainValidationOptions", [])
                if opt.get("ValidationMethod") == "DNS" and "ResourceRecord" in opt
            ]
        LOGGER.info(
            "removing %i validation record(s) from hosted zone: %s",
            len(records),
            self.args.hosted_zone_id,
        )
        self.__change_record_set("DELETE", records)

    def update_record_set(self, record_set: ResourceRecordTypeDef) -> None:
        """Update a validation record set when the cert has not changed.

        Args:
            record_set: Record set to be updated in Route 53.

        """
        LOGGER.info("updating record set...")
        self.__change_record_set("UPSERT", [record_set])

    def __change_record_set(
        self,
        action: Literal["CREATE", "DELETE", "UPSERT"],
        record_sets: List[ResourceRecordTypeDef],
    ) -> None:
        """Wrap boto3.client('acm').change_resource_record_sets.

        Args:
            action: Change action.
            record_sets: Record sets to change.

        """
        if not record_sets:
            raise ValueError("Must provide one of more record sets")

        changes: List[ChangeTypeDef] = [
            {
                "Action": action,
                "ResourceRecordSet": {
                    "Name": record["Name"],
                    "Type": record["Type"],
                    "ResourceRecords": [{"Value": record["Value"]}],
                    "TTL": self.args.ttl,
                },
            }
            for record in record_sets
        ]

        LOGGER.debug(
            'making the following changes to hosted zone "%s":\n%s',
            self.args.hosted_zone_id,
            changes,
        )

        self.r53_client.change_resource_record_sets(
            HostedZoneId=self.args.hosted_zone_id,
            ChangeBatch={"Comment": self.template_description, "Changes": changes},
        )

    def deploy(self, status: Optional[Status] = None) -> Dict[str, str]:
        """Deploy an ACM Certificate."""
        record = None
        try:
            if self.domain_changed():
                return {}

            if not status:
                status = self.deploy_stack()

            cert_arn = self.get_certificate()
            result = {"CertificateArn": cert_arn}

            if status == NO_CHANGE:
                LOGGER.debug("stack did not change; no action required")
                return result

            if status == SUBMITTED:
                if status.reason == "creating new stack":
                    record = self.get_validation_record(cert_arn)
                    self.put_record_set(record)
                    LOGGER.info(
                        "waiting for certificate to validate; "
                        "this can take upwards of 30 minutes to "
                        "complete..."
                    )
                elif status.reason == "updating existing stack":
                    # get the cert ARN again in case it changed during update
                    cert_arn = self.get_certificate()
                    record = self.get_validation_record(cert_arn, status="SUCCESS")
                    self.update_record_set(record)
                elif status.reason == "destroying stack for re-creation":
                    # handle recreating a stack in a failed state
                    return self.deploy(
                        status=self._wait_for_stack(
                            self._deploy_action,
                            last_status=status,
                            till_reason="creating new stack",
                        )
                    )
                self._wait_for_stack(self._deploy_action, last_status=status)
                return result
        except (
            self.r53_client.exceptions.InvalidChangeBatch,
            self.r53_client.exceptions.NoSuchHostedZone,
            StackFailed,
        ) as err:
            LOGGER.error(err)
            self.destroy(
                records=[record] if record else None,
                skip_r53=isinstance(  # type: ignore
                    err,
                    (
                        self.r53_client.exceptions.InvalidChangeBatch,
                        self.r53_client.exceptions.NoSuchHostedZone,
                    ),
                ),
            )
        except StackUpdateBadStatus as err:
            # don't try to destroy the stack when it can be in progress
            LOGGER.error(err)
        return {}

    def destroy(
        self,
        records: Optional[List[ResourceRecordTypeDef]] = None,
        skip_r53: bool = False,
    ) -> bool:
        """Destroy an ACM certificate.

        Args:
            records: List of validation records to remove from Route 53.
                This can be provided in cases were the certificate has been
                deleted during a rollback.
            skip_r53: Skip the removal of validation records.

        """
        if not skip_r53:
            try:
                self.remove_validation_records(records)
            except (
                self.r53_client.exceptions.InvalidChangeBatch,
                self.r53_client.exceptions.NoSuchHostedZone,
                self.acm_client.exceptions.ResourceNotFoundException,
            ) as err:
                # these error are fine if they happen during destruction but
                # could require manual steps to finish cleanup.
                LOGGER.warning(
                    "deletion of the validation records failed with error:\n%s", err
                )
            except ClientError as err:
                if err.response["Error"]["Message"] != (
                    f"Stack with id {self.stack.fqn} does not exist"
                ):
                    raise
                LOGGER.warning(
                    "deletion of the validation records failed with error:\n%s", err
                )
        else:
            LOGGER.info("deletion of validation records was skipped")
        self.destroy_stack(wait=True)
        return True

    def post_deploy(self) -> Dict[str, str]:
        """Run during the **post_deploy** stage."""
        return self.deploy()

    def post_destroy(self) -> bool:
        """Run during the **post_destroy** stage."""
        return self.destroy()

    def pre_deploy(self) -> Dict[str, str]:
        """Run during the **pre_deploy** stage."""
        return self.deploy()

    def pre_destroy(self) -> bool:
        """Run during the **pre_destroy** stage."""
        return self.destroy()



"""``runway deploy`` command."""
# docs: file://./../../../docs/source/commands.rst
import logging
from typing import Any, Tuple

import click
from pydantic import ValidationError

from ...core import Runway
from ...exceptions import ConfigNotFound, VariablesFileNotFound
from .. import options
from ..utils import select_deployments

LOGGER = logging.getLogger(__name__.replace("._", "."))


@click.command("deploy", short_help="deploy things")
@options.ci
@options.debug
@options.deploy_environment
@options.no_color
@options.tags
@options.verbose
@click.pass_context
def deploy(ctx: click.Context, debug: bool, tags: Tuple[str, ...], **_: Any) -> None:
    """Deploy infrastructure as code.

    \b
    Process
    -------
    1. Determines the deploy environment.
        - "-e, --deploy-environment" option
        - "DEPLOY_ENVIRONMENT" environment variable
        - git branch name
            - strips "ENV-" prefix, master is converted to common
            - ignored if "ignore_git_branch: true"
        - name of the current working directory
    2. Selects deployments & modules to deploy.
        - (default) prompts
        - (tags) module contains all tags
        - (non-interactive) all
    3. Deploys selected deployments/modules in the order defined.

    """  # noqa: D301
    try:
        Runway(ctx.obj.runway_config, ctx.obj.get_runway_context()).deploy(
            select_deployments(ctx, ctx.obj.runway_config.deployments, tags)
        )
    except ValidationError as err:
        LOGGER.error(err, exc_info=debug)
        ctx.exit(1)
    except (ConfigNotFound, VariablesFileNotFound) as err:
        LOGGER.error(err.message, exc_info=debug)
        ctx.exit(1)



"""Poetry interface."""
from __future__ import annotations

import logging
import re
import subprocess
from pathlib import Path
from typing import TYPE_CHECKING, Any, List, Optional, Tuple

import tomli
from typing_extensions import Final, Literal

from ..compat import cached_property
from ..exceptions import RunwayError
from ..utils import Version
from .base_classes import DependencyManager

if TYPE_CHECKING:
    from _typeshed import StrPath

LOGGER = logging.getLogger(__name__)


class PoetryExportFailedError(RunwayError):
    """Poetry export failed to produce a ``requirements.txt`` file."""

    def __init__(self, output: str, *args: Any, **kwargs: Any) -> None:
        """Instantiate class. All args/kwargs are passed to parent method.

        Args:
            output: The output from running ``poetry export``.

        """
        self.message = f"poetry export failed with the following output:\n{output}"
        super().__init__(*args, **kwargs)


class PoetryNotFoundError(RunwayError):
    """Poetry not installed or found in $PATH."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Instantiate class. All args/kwargs are passed to parent method."""
        self.message = (
            "poetry not installed or not in PATH! "
            "Install it according to poetry docs (https://python-poetry.org/) "
            "and ensure it is available in PATH."
        )
        super().__init__(*args, **kwargs)


class Poetry(DependencyManager):
    """Poetry dependency manager."""

    CONFIG_FILES: Final[Tuple[Literal["poetry.lock"], Literal["pyproject.toml"]]] = (
        "poetry.lock",
        "pyproject.toml",
    )
    """Configuration files used by poetry."""

    EXECUTABLE: Final[Literal["poetry"]] = "poetry"
    """CLI executable."""

    @cached_property
    def version(self) -> Version:
        """poetry version."""
        cmd_output = self._run_command([self.EXECUTABLE, "--version"])
        match = re.search(r"^Poetry version (?P<version>\S*)", cmd_output)
        if not match:
            LOGGER.warning(
                "unable to parse poetry version from output:\n%s", cmd_output
            )
            return Version("0.0.0")
        return Version(match.group("version"))

    @classmethod
    def dir_is_project(cls, directory: StrPath, **__kwargs: Any) -> bool:
        """Determine if the directory contains a project for this dependency manager.

        Args:
            directory: Directory to check.

        """
        pyproject_path = Path(directory) / Poetry.CONFIG_FILES[1]

        if not pyproject_path.is_file():
            return False

        # check for PEP-517 definition
        pyproject = tomli.loads(pyproject_path.read_text())
        build_system_requires: Optional[List[str]] = pyproject.get(
            "build-system", {}
        ).get("requires")

        if build_system_requires:
            for req in build_system_requires:
                if req.startswith("poetry"):
                    LOGGER.debug("poetry project detected")
                    return True
        return False

    def export(
        self,
        *,
        dev: bool = False,
        extras: Optional[List[str]] = None,
        output: StrPath,
        output_format: str = "requirements.txt",
        with_credentials: bool = True,
        without_hashes: bool = True,
    ) -> Path:
        """Export the lock file to other formats.

        Args:
            dev: Include development dependencies.
            extras: Extra sets of dependencies to include.
            output: Path to the output file.
            output_format: The format to export to.
            with_credentials: Include credentials for extra indices.
            without_hashes: Exclude hashes from the exported file.

        Returns:
            Path to the output file.

        """
        output = Path(output)
        try:
            result = self._run_command(
                self.generate_command(
                    "export",
                    dev=dev,
                    extras=extras,
                    format=output_format,
                    output=output.name,
                    with_credentials=with_credentials,
                    without_hashes=without_hashes,
                )
            )
            requirements_txt = self.cwd / output.name
            if requirements_txt.is_file():
                output.parent.mkdir(exist_ok=True, parents=True)
                requirements_txt.rename(output)  # python3.7 doesn't return the new path
                return output
        except subprocess.CalledProcessError as exc:
            raise PoetryExportFailedError(exc.stderr) from exc
        raise PoetryExportFailedError(result)



"""Register sync strategies.

.. Derived from software distributed by Amazon.com, Inc - http://aws.amazon.com/apache2.0/
   https://github.com/aws/aws-cli/blob/83b43782dd/awscli/customizations/s3/syncstrategy/register.py

"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Type

from .delete import DeleteSync
from .exact_timestamps import ExactTimestampsSync
from .size_only import SizeOnlySync

if TYPE_CHECKING:
    from botocore.session import Session

    from .base import BaseSync, ValidSyncType


def register_sync_strategy(
    session: Session,
    strategy_cls: Type[BaseSync],
    sync_type: ValidSyncType = "file_at_src_and_dest",
):
    """Register a single sync strategy.

    Args:
        session: The session that the sync strategy is being registered to.
        strategy_cls: The class of the sync strategy to be registered.
        sync_type: A string representing when to perform the sync strategy.
            See ``__init__`` method of ``BaseSyncStrategy`` for possible options.

    """
    strategy = strategy_cls(sync_type)
    strategy.register_strategy(session)


def register_sync_strategies(session: Session, **_: Any) -> None:
    """Register the different sync strategies."""
    # Register the size only sync strategy.
    register_sync_strategy(session, SizeOnlySync)

    # Register the exact timestamps sync strategy.
    register_sync_strategy(session, ExactTimestampsSync)

    # Register the delete sync strategy.
    register_sync_strategy(session, DeleteSync, "file_not_at_src")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['runway',
 'runway._cli',
 'runway._cli.commands',
 'runway._cli.commands._gen_sample',
 'runway._cli.commands._kbenv',
 'runway._cli.commands._schema',
 'runway._cli.commands._tfenv',
 'runway.aws_sso_botocore',
 'runway.blueprints',
 'runway.blueprints.k8s',
 'runway.blueprints.staticsite',
 'runway.cfngin',
 'runway.cfngin.actions',
 'runway.cfngin.blueprints',
 'runway.cfngin.blueprints.variables',
 'runway.cfngin.dag',
 'runway.cfngin.hooks',
 'runway.cfngin.hooks.awslambda',
 'runway.cfngin.hooks.awslambda.models',
 'runway.cfngin.hooks.awslambda.python_requirements',
 'runway.cfngin.hooks.docker',
 'runway.cfngin.hooks.docker.image',
 'runway.cfngin.hooks.ecr',
 'runway.cfngin.hooks.ssm',
 'runway.cfngin.hooks.staticsite',
 'runway.cfngin.hooks.staticsite.auth_at_edge',
 'runway.cfngin.hooks.staticsite.auth_at_edge.templates',
 'runway.cfngin.hooks.staticsite.auth_at_edge.templates.check_auth',
 'runway.cfngin.hooks.staticsite.auth_at_edge.templates.http_headers',
 'runway.cfngin.hooks.staticsite.auth_at_edge.templates.parse_auth',
 'runway.cfngin.hooks.staticsite.auth_at_edge.templates.refresh_auth',
 'runway.cfngin.hooks.staticsite.auth_at_edge.templates.sign_out',
 'runway.cfngin.logger',
 'runway.cfngin.lookups',
 'runway.cfngin.lookups.handlers',
 'runway.cfngin.providers',
 'runway.cfngin.providers.aws',
 'runway.config',
 'runway.config.components',
 'runway.config.components.runway',
 'runway.config.models',
 'runway.config.models.cfngin',
 'runway.config.models.runway',
 'runway.config.models.runway.options',
 'runway.context',
 'runway.core',
 'runway.core.components',
 'runway.core.providers',
 'runway.core.providers.aws',
 'runway.core.providers.aws.s3',
 'runway.core.providers.aws.s3._helpers',
 'runway.core.providers.aws.s3._helpers.sync_strategy',
 'runway.dependency_managers',
 'runway.env_mgr',
 'runway.lookups',
 'runway.lookups.handlers',
 'runway.module',
 'runway.module.staticsite',
 'runway.module.staticsite.options',
 'runway.module.staticsite.parameters',
 'runway.sources',
 'runway.templates.cdk-py',
 'runway.templates.cdk-py.hello',
 'runway.templates.k8s-cfn-repo.k8s-master.cfn.k8s_hooks',
 'runway.templates.k8s-cfn-repo.k8s-workers.cfn.local_lookups',
 'runway.templates.k8s-flux-repo.flux.tf',
 'runway.templates.k8s-tf-repo.eks-base.tf',
 'runway.templates.k8s-tf-repo.gen-kubeconfig.cfn.k8s_hooks',
 'runway.templates.k8s-tf-repo.job-s3-echo.tf',
 'runway.templates.sls-py',
 'runway.templates.sls-py.hello_world',
 'runway.tests',
 'runway.tests.handlers',
 'runway.utils']

package_data = \
{'': ['*'],
 'runway': ['templates/*',
            'templates/cdk-csharp/*',
            'templates/cdk-csharp/src/*',
            'templates/cdk-csharp/src/HelloCdk/*',
            'templates/cdk-tsc/*',
            'templates/cdk-tsc/bin/*',
            'templates/cdk-tsc/lib/*',
            'templates/cfn/*',
            'templates/cfngin/*',
            'templates/k8s-cfn-repo/*',
            'templates/k8s-cfn-repo/aws-auth-cm.k8s/base/*',
            'templates/k8s-cfn-repo/aws-auth-cm.k8s/overlays/template/*',
            'templates/k8s-cfn-repo/k8s-master.cfn/*',
            'templates/k8s-cfn-repo/k8s-workers.cfn/*',
            'templates/k8s-cfn-repo/service-hello-world.k8s/*',
            'templates/k8s-cfn-repo/service-hello-world.k8s/base/*',
            'templates/k8s-cfn-repo/service-hello-world.k8s/overlays/prod/*',
            'templates/k8s-cfn-repo/service-hello-world.k8s/overlays/template/*',
            'templates/k8s-flux-repo/*',
            'templates/k8s-flux-repo/flux-dev/*',
            'templates/k8s-flux-repo/flux-dev/namespaces/*',
            'templates/k8s-flux-repo/flux-dev/workloads/*',
            'templates/k8s-tf-repo/*',
            'templates/k8s-tf-repo/gen-kubeconfig.cfn/*',
            'templates/k8s-tf-repo/service-hello-world.k8s/*',
            'templates/k8s-tf-repo/service-hello-world.k8s/base/*',
            'templates/k8s-tf-repo/service-hello-world.k8s/overlays/dev/*',
            'templates/k8s-tf-repo/service-hello-world.k8s/overlays/prod/*',
            'templates/k8s-tf-repo/tfstate.cfn/*',
            'templates/sls-tsc/*',
            'templates/sls-tsc/src/*',
            'templates/static-angular/*',
            'templates/static-angular/sampleapp.web/*',
            'templates/static-angular/sampleapp.web/e2e/*',
            'templates/static-angular/sampleapp.web/e2e/src/*',
            'templates/static-angular/sampleapp.web/src/*',
            'templates/static-angular/sampleapp.web/src/app/*',
            'templates/static-angular/sampleapp.web/src/assets/*',
            'templates/static-angular/sampleapp.web/src/environments/*',
            'templates/static-react/*',
            'templates/static-react/sampleapp.web/*',
            'templates/static-react/sampleapp.web/public/*',
            'templates/static-react/sampleapp.web/src/*',
            'templates/terraform/backend-us-east-1.tfvars',
            'templates/terraform/backend-us-east-1.tfvars',
            'templates/terraform/backend-us-east-1.tfvars',
            'templates/terraform/dev-us-east-1.tfvars',
            'templates/terraform/dev-us-east-1.tfvars',
            'templates/terraform/dev-us-east-1.tfvars',
            'templates/terraform/main.tf',
            'templates/terraform/main.tf',
            'templates/terraform/main.tf'],
 'runway.blueprints.staticsite': ['templates/*'],
 'runway.templates.k8s-flux-repo.flux.tf': ['docker/*']}

install_requires = \
['awacs',
 'boto3>=1.16,<2.0',
 'cfn-lint',
 'cfn_flip>=1.2,<2.0',
 'click>=8.0,<9.0',
 'coloredlogs!=15.0.1',
 'docker>=3.0.0',
 'formic2',
 'gitpython',
 'igittigitt>=2.0.5',
 'jinja2>=2.7',
 'moto==3.0.5',
 'packaging',
 'pipenv==2022.1.8',
 'pyOpenSSL',
 'pydantic>=1.4,<2.0',
 'pyhcl>=0.4,<0.5',
 'python-hcl2>=3.0.0',
 'pyyaml>5.4',
 'requests',
 'send2trash',
 'str2bool>=1.1,<2.0',
 'testfixtures>=7.0.3,<8.0.0',
 'tomli>=1.2.2',
 'troposphere>=2.4,<4',
 'typing_extensions',
 'urllib3',
 'yamllint']

extras_require = \
{':python_version < "3.8"': ['backports.cached_property', 'importlib-metadata'],
 'docs': ['dunamai>=1.5,<2.0',
          'jsx-lexer>=1.0,<2.0',
          'sphinx>=4.3,<5.0',
          'sphinx-github-changelog>=1.1,<2.0',
          'sphinx-rtd-theme>=1.0,<2.0',
          'sphinx-tabs>=3.2,<4.0',
          'sphinxcontrib-apidoc>=0.3,<0.4',
          'sphinxcontrib-programoutput>=0.17,<0.18']}

entry_points = \
{'console_scripts': ['runway = runway._cli.main:cli']}

setup_kwargs = {
    'name': 'runway',
    'version': '2.6.4',
    'description': 'Simplify infrastructure/app testing/deployment',
    'long_description': '# Runway\n\n[![CI/CD](https://github.com/onicagroup/runway/workflows/CI/CD/badge.svg?branch=master)](https://github.com/onicagroup/runway/actions?query=workflow%3ACI%2FCD)\n[![codecov](https://codecov.io/gh/onicagroup/runway/branch/master/graph/badge.svg?token=Ku28I0RY80)](https://codecov.io/gh/onicagroup/runway)\n[![PyPi](https://img.shields.io/pypi/v/runway?style=flat)](https://pypi.org/project/runway/)\n[![npm](https://img.shields.io/npm/v/@onica/runway?style=flat)](https://www.npmjs.com/package/@onica/runway)\n[![code style: black](https://img.shields.io/badge/code%20style-black-000000.svg?style=flat)](https://github.com/psf/black)\n\n![runway-example.gif](https://raw.githubusercontent.com/onicagroup/runway/master/docs/source/images/runway-example.gif)\n\nRunway is a lightweight integration app designed to ease management of infrastructure tools.\n\nIts main goals are to encourage GitOps best-practices, avoid convoluted Makefiles/scripts (enabling identical deployments from a workstation or CI job), and enable developers/admins to use the best tool for any given job.\n\n\n## Features\n\n* Centralized environment-specific configuration\n* Automatic environment identification from git branches\n* Automatic linting/verification\n* Support of IAM roles to assume for each deployment\n* Terraform backend/workspace config management w/per-environment tfvars\n* Automatic kubectl/terraform version management per-environment\n\n### Supported Deployment Tools\n\n* AWS CDK\n* Kubectl\n* Serverless Framework\n* CFNgin (CloudFormation)\n* Static websites (build & deploy to S3+CloudFront)\n* Terraform\n\n\n## Example\n\nA typical Runway configuration is unobtrusive -- it just lists the deployment order and locations (regions).\n\n```yml\ndeployments:\n  - modules:\n      - resources.tf  # terraform resources\n      - backend.sls  # serverless lambda functions\n      - frontend  # static web site\n    environments:  # Environments\n      dev: "123456789012"  # AWS development Account ID\n      prod: "234567890123"  # AWS production Account ID\n    regions:\n      - us-east-1\n```\n\nThe example above contains enough information for Runway to deploy all resources, lambda functions and a static website backed by S3 and Cloudfront in either dev or prod environments\n\n\n## Install\n\nRunway is available via any of the following installation methods. Use whatever works best for your project/team (it\'s the same application no matter how you obtain it).\n\n### HTTPS Download (e.g cURL)\n\nUse one of the endpoints below to download a single-binary executable version of Runway based on your operating system.\n\n| Operating System | Endpoint                               |\n|------------------|----------------------------------------|\n| Linux            | <https://oni.ca/runway/latest/linux>   |\n| macOS            | <https://oni.ca/runway/latest/osx>     |\n| Windows          | <https://oni.ca/runway/latest/windows> |\n\n```shell\n$ curl -L oni.ca/runway/latest/osx -o runway\n$ chmod +x runway\n$ ./runway new\n```\n\n**Suggested use:** CloudFormation or Terraform projects\n\n\n### npm\n\n```shell\n$ npm i -D @onica/runway\n$ npx runway new\n```\n\n**Suggested use:** Serverless or AWS CDK projects\n\n\n### pip (or poetry, etc)\n\n```shell\n$ pip install runway\n$ runway new\n# OR\n$ poetry add --dev runway\n$ poetry run runway new\n```\n\n**Suggested use:** Python projects\n\n\n## Documentation\n\nSee the [doc site](https://docs.onica.com/projects/runway) for full documentation.\n\nQuickstart documentation, including CloudFormation templates and walkthrough can be found [here](https://docs.onica.com/projects/runway/page/quickstart/index.html)\n',
    'author': 'Onica Group LLC',
    'author_email': 'opensource@onica.com',
    'maintainer': 'Kyle Finley',
    'maintainer_email': 'kyle@finley.sh',
    'url': 'https://github.com/onicagroup/runway',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)



"""Hook utils."""
from __future__ import annotations

import collections.abc
import logging
import os
import sys
from typing import TYPE_CHECKING, Any, Dict, List, cast

import pydantic

from ...exceptions import FailedVariableLookup
from ...utils import BaseModel, load_object_from_string
from ...variables import Variable, resolve_variables
from ..blueprints.base import Blueprint

if TYPE_CHECKING:
    from ...config.models.cfngin import CfnginHookDefinitionModel
    from ...context import CfnginContext
    from ..providers.aws.default import Provider

LOGGER = logging.getLogger(__name__)


class BlankBlueprint(Blueprint):
    """Blueprint that can be built programatically."""

    def create_template(self) -> None:
        """Create template without raising NotImplementedError."""


# TODO BREAKING find a better place for this
class TagDataModel(BaseModel):
    """AWS Resource Tag data model."""

    key: str
    value: str

    class Config:
        """Model configuration."""

        allow_population_by_field_name = True
        extra = pydantic.Extra.forbid
        fields = {
            "key": {"alias": "Key"},
            "value": {"alias": "Value"},
        }


def full_path(path: str) -> str:
    """Return full path."""
    return os.path.abspath(os.path.expanduser(path))


# TODO split up to reduce number of statements
def handle_hooks(  # pylint: disable=too-many-statements
    stage: str,
    hooks: List[CfnginHookDefinitionModel],
    provider: Provider,
    context: CfnginContext,
):
    """Handle pre/post_deploy hooks.

    These are pieces of code that we want to run before/after deploying
    stacks.

    Args:
        stage: The current stage (pre_run, post_run, etc).
        hooks: Hooks to execute.
        provider: Provider instance.
        context: Context instance.

    """
    if not hooks:
        LOGGER.debug("no %s hooks defined", stage)
        return

    hook_paths: List[str] = []
    for i, hook in enumerate(hooks):
        try:
            hook_paths.append(hook.path)
        except KeyError as exc:
            raise ValueError(f"{stage} hook #{i} missing path.") from exc

    LOGGER.info("executing %s hooks: %s", stage, ", ".join(hook_paths))
    for hook in hooks:
        if not hook.enabled:
            LOGGER.debug("hook with method %s is disabled; skipping", hook.path)
            continue

        try:
            method = load_object_from_string(hook.path, try_reload=True)
        except (AttributeError, ImportError):
            LOGGER.exception("unable to load method at %s", hook.path)
            if hook.required:
                raise
            continue

        if hook.args:
            args = [Variable(k, v) for k, v in hook.args.items()]
            try:  # handling for output or similar being used in pre_deploy
                resolve_variables(args, context, provider)
            except FailedVariableLookup:
                if "pre" in stage:
                    LOGGER.error(
                        "lookups that change the order of execution, like "
                        '"output", can only be used in "post_*" hooks; '
                        "please ensure that the hook being used does "
                        "not rely on a stack, hook_data, or context that "
                        "does not exist yet"
                    )
                raise
            kwargs: Dict[str, Any] = {v.name: v.value for v in args}
        else:
            kwargs = {}

        try:
            if isinstance(method, type):
                result: Any = getattr(
                    method(context=context, provider=provider, **kwargs), stage
                )()
            else:
                result = cast(Any, method(context=context, provider=provider, **kwargs))
        except Exception:  # pylint: disable=broad-except
            LOGGER.exception("hook %s threw an exception", hook.path)
            if hook.required:
                raise
            continue

        if not result:
            if hook.required:
                LOGGER.error(
                    "required hook %s failed; return value: %s", hook.path, result
                )
                sys.exit(1)
            LOGGER.warning(
                "non-required hook %s failed; return value: %s", hook.path, result
            )
        else:
            if isinstance(result, (collections.abc.Mapping, pydantic.BaseModel)):
                if hook.data_key:
                    LOGGER.debug(
                        "adding result for hook %s to context in data_key %s",
                        hook.path,
                        hook.data_key,
                    )
                    context.set_hook_data(hook.data_key, result)
                else:
                    LOGGER.debug(
                        "hook %s returned result data but no data key set; ignoring",
                        hook.path,
                    )



"""``runway test`` command."""
# docs: file://./../../../docs/source/commands.rst
import logging
from typing import Any

import click

from ...core import Runway
from .. import options

LOGGER = logging.getLogger(__name__.replace("._", "."))


@click.command("test", short_help="run tests")
@options.debug
@options.deploy_environment
@options.no_color
@options.verbose
@click.pass_context
def test(ctx: click.Context, **_: Any) -> None:
    """Execute tests as defined in the Runway config file.

    If one of the tests fail, the command will exit immediately unless
    "required: false" is set on the failing test.

    If the failing test is not required, the next test will be executed.

    If any of the tests fail, the command will exit with a non-zero exit code.

    """
    Runway(ctx.obj.runway_config, ctx.obj.get_runway_context()).test()



"""Runway deployment object."""
from __future__ import annotations

import concurrent.futures
import logging
import multiprocessing
import sys
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from ..._logging import PrefixAdaptor
from ...compat import cached_property
from ...config.components.runway import RunwayVariablesDefinition
from ...config.models.runway import (
    RunwayAssumeRoleDefinitionModel,
    RunwayFutureDefinitionModel,
)
from ...exceptions import UnresolvedVariable
from ...utils import flatten_path_lists, merge_dicts
from ..providers import aws
from ._module import Module

if TYPE_CHECKING:
    from ...config.components.runway import RunwayDeploymentDefinition
    from ...context import RunwayContext
    from ..type_defs import RunwayActionTypeDef


LOGGER = logging.getLogger(__name__.replace("._", "."))


class Deployment:
    """Runway deployment."""

    def __init__(
        self,
        context: RunwayContext,
        definition: RunwayDeploymentDefinition,
        future: Optional[RunwayFutureDefinitionModel] = None,
        variables: Optional[RunwayVariablesDefinition] = None,
    ) -> None:
        """Instantiate class.

        Args:
            context: Runway context object.
            definition: A single deployment definition.
            future: Future functionality configuration.
            variables: Runway variables.

        """
        self._future = future or RunwayFutureDefinitionModel()
        self._variables = variables or RunwayVariablesDefinition.parse_obj({})
        self.definition = definition
        self.ctx = context
        self.name = self.definition.name
        self.logger = PrefixAdaptor(self.name, LOGGER)
        self.__merge_env_vars()

    @property
    def assume_role_config(self) -> Dict[str, Union[bool, int, str]]:
        """Parse the definition to get assume role arguments."""
        assume_role = self.definition.assume_role
        if not assume_role:
            self.logger.debug(
                "assume_role not configured for deployment: %s", self.name
            )
            return {}
        if isinstance(assume_role, str):  # type: ignore
            self.logger.debug("role found: %s", assume_role)
            assume_role = RunwayAssumeRoleDefinitionModel(arn=assume_role)
        elif isinstance(assume_role, dict):  # type: ignore
            assume_role = RunwayAssumeRoleDefinitionModel.parse_obj(assume_role)
        if not assume_role.arn:
            self.logger.debug(
                "assume_role not configured for deployment: %s", self.name
            )
            return {}
        return {
            "duration_seconds": assume_role.duration,
            "revert_on_exit": assume_role.post_deploy_env_revert,
            "role_arn": assume_role.arn,
            "session_name": assume_role.session_name,
        }

    @property
    def env_vars_config(self) -> Dict[str, str]:
        """Parse the definition to get the correct env_vars configuration."""
        try:
            if not self.definition.env_vars:
                return {}
        except UnresolvedVariable:
            # pylint: disable=protected-access
            if "env_vars" in self.definition._vars:
                var = self.definition._vars["env_vars"]
                var.resolve(self.ctx, variables=self._variables)
                self.definition._data["env_vars"] = var.value
            else:
                raise
        return flatten_path_lists(self.definition.env_vars, str(self.ctx.env.root_dir))

    @cached_property
    def regions(self) -> List[str]:
        """List of regions this deployment is associated with."""
        return self.definition.parallel_regions or self.definition.regions

    @cached_property
    def use_async(self) -> bool:
        """Whether to use asynchronous method."""
        return bool(self.definition.parallel_regions and self.ctx.use_concurrent)

    def deploy(self) -> None:
        """Deploy the deployment.

        High level method for running a deployment.

        """
        self.logger.verbose(
            "attempting to deploy to region(s): %s", ", ".join(self.regions)
        )
        if self.use_async:
            return self.__async("deploy")
        return self.__sync("deploy")

    def destroy(self) -> None:
        """Destroy the deployment.

        High level method for running a deployment.

        """
        self.logger.verbose(
            "attempting to destroy in region(s): %s", ", ".join(self.regions)
        )
        if self.use_async:
            return self.__async("destroy")
        return self.__sync("destroy")

    def init(self) -> None:
        """Initialize/bootstrap deployment.

        High level method for running a deployment.

        """
        self.logger.verbose(
            "attempting to initialize region(s): %s", ", ".join(self.regions)
        )
        if self.use_async:
            return self.__async("init")
        return self.__sync("init")

    def plan(self) -> None:
        """Plan for the next deploy of the deployment.

        High level method for running a deployment.

        """
        self.logger.verbose(
            "attempting to plan for the next deploy to region(s): %s",
            ", ".join(self.regions),
        )
        if self.use_async:
            self.logger.notice(
                "processing of regions will be done in parallel during deploy/destroy"
            )
        return self.__sync("plan")

    def run(self, action: RunwayActionTypeDef, region: str) -> None:
        """Run a single deployment in a single region.

        Low level API access to run a deployment object.

        Args:
            action: Action to run (deploy, destroy, plan, etc.)
            region: AWS region to run in.

        """
        context = self.ctx.copy() if self.use_async else self.ctx
        context.command = action
        context.env.aws_region = region

        with aws.AssumeRole(context, **self.assume_role_config):
            self.definition.resolve(context, variables=self._variables)
            self.validate_account_credentials(context)
            Module.run_list(
                action=action,
                context=context,
                deployment=self.definition,
                future=self._future,
                modules=self.definition.modules,
                variables=self._variables,
            )

    def validate_account_credentials(
        self, context: Optional[RunwayContext] = None
    ) -> None:
        """Exit if requested deployment account doesn't match credentials.

        Args:
            context: Context object.

        Raises:
            SystemExit: AWS Account associated with the current credentials
                did not match the defined criteria.

        """
        account = aws.AccountDetails(context or self.ctx)
        if self.definition.account_id:
            if self.definition.account_id != account.id:
                self.logger.error(
                    'current AWS account "%s" does not match '
                    'required account "%s" in Runway config',
                    account.id,
                    self.definition.account_id,
                )
                sys.exit(1)
            self.logger.info(
                "verified current AWS account matches required " 'account id "%s"',
                self.definition.account_id,
            )
        if self.definition.account_alias:
            if self.definition.account_alias not in account.aliases:
                self.logger.error(
                    'current AWS account aliases "%s" do not match '
                    'required account alias "%s" in Runway config.',
                    ",".join(account.aliases),
                    self.definition.account_alias,
                )
                sys.exit(1)
            self.logger.info(
                'verified current AWS account alias matches required alias "%s"',
                self.definition.account_alias,
            )

    def __merge_env_vars(self) -> None:
        """Merge defined env_vars into context.env_vars."""
        if self.env_vars_config:
            self.logger.verbose(
                "environment variable overrides are being applied to this deployment"
            )
            self.logger.debug(
                "environment variable overrides: %s", self.env_vars_config
            )
            self.ctx.env.vars = merge_dicts(self.ctx.env.vars, self.env_vars_config)

    def __async(self, action: RunwayActionTypeDef) -> None:
        """Execute asynchronously.

        Args:
            action: Name of action to run.

        """
        self.logger.info(
            "processing regions in parallel... (output will be interwoven)"
        )
        with concurrent.futures.ProcessPoolExecutor(
            max_workers=self.ctx.env.max_concurrent_regions,
            mp_context=multiprocessing.get_context("fork"),
        ) as executor:
            futures = [
                executor.submit(self.run, *[action, region]) for region in self.regions
            ]
        for job in futures:
            job.result()  # raise exceptions / exit as needed

    def __sync(self, action: RunwayActionTypeDef) -> None:
        """Execute synchronously.

        Args:
            action: Name of action to run.

        """
        self.logger.info("processing regions sequentially...")
        for region in self.regions:
            self.logger.verbose("processing AWS region: %s", region)
            self.run(action, region)

    @classmethod
    def run_list(
        cls,
        action: RunwayActionTypeDef,
        context: RunwayContext,
        deployments: List[RunwayDeploymentDefinition],
        future: RunwayFutureDefinitionModel,
        variables: RunwayVariablesDefinition,
    ) -> None:
        """Run a list of deployments.

        Args:
            action: Name of action to run.
            context: Runway context.
            deployments: List of deployments to run.
            future: Future definition.
            variables: Runway variables for lookup resolution.

        """
        for definition in deployments:
            definition.resolve(context, variables=variables, pre_process=True)
            deployment = cls(
                context=context,
                definition=definition,
                future=future,
                variables=variables,
            )
            LOGGER.info("")
            LOGGER.info("")
            deployment.logger.notice("processing deployment (in progress)")
            if not definition.modules:
                deployment.logger.warning("skipped; no modules found in definition")
                continue
            cls(
                context=context,
                definition=definition,
                future=future,
                variables=variables,
            )[action]()
            deployment.logger.success("processing deployment (complete)")

    def __getitem__(self, name: str) -> Any:
        """Make the object subscriptable.

        Args:
            name: Attribute to get.

        """
        return getattr(self, name)



"""Click options."""
import click

ci = click.option(
    "--ci",
    default=False,
    envvar="CI",
    is_flag=True,
    help="Run in non-interactive mode.",
)

debug = click.option(
    "--debug",
    count=True,
    envvar="DEBUG",
    help="Supply once to display Runway debug logs. "
    "Supply twice to display all debug logs.",
)

deploy_environment = click.option(
    "-e",
    "--deploy-environment",
    envvar="DEPLOY_ENVIRONMENT",
    metavar="<env-name>",
    help="Manually specify the name of the deploy environment.",
)

no_color = click.option(
    "--no-color",
    default=False,
    envvar="RUNWAY_NO_COLOR",
    is_flag=True,
    help="Disable color in Runway's logs.",
)

tags = click.option(
    "--tag",
    "tags",
    metavar="<tag>...",
    multiple=True,
    help="Select modules by tag or tags. "
    "This option can be specified more than once to build a"
    ' list of tags that are treated as "AND". '
    '(e.g. "--tag <tag1> --tag <tag2>" would select all modules'
    " with BOTH tags).",
)

verbose = click.option(
    "--verbose",
    default=False,
    envvar="VERBOSE",
    is_flag=True,
    help="Display Runway verbose logs.",
)



"""Static site utilities."""


def add_url_scheme(url: str) -> str:
    """Add the scheme to an existing url.

    Args:
        url (str): The current url.

    """
    if url.startswith("https://") or url.startswith("http://"):
        return url
    return f"https://{url}"
